"""GeckoConfigStruct - A class to manage the ConfigStruct for 'inClear-32K v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 2

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "inClear-32K-Mode": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Mode",
                512,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "inClear-32K-MaintenanceLevel": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-MaintenanceLevel",
                513,
                "ALL",
            ),
            "inClear-32K-BoostLevel": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/All/inClear-32K-BoostLevel", 514, "ALL"
            ),
            "inClear-32K-MaxCellCurrent": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-MaxCellCurrent",
                515,
                "ALL",
            ),
            "inClear-32K-MaxMaintenanceLevel": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-MaxMaintenanceLevel",
                517,
                "ALL",
            ),
            "inClear-32K-ErrDelayAfterReset": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-ErrDelayAfterReset",
                519,
                "ALL",
            ),
            "inClear-32K-Boost1Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost1Durx6Minutes",
                521,
                "ALL",
            ),
            "inClear-32K-Boost2Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost2Durx6Minutes",
                523,
                "ALL",
            ),
            "inClear-32K-Boost3Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost3Durx6Minutes",
                525,
                "ALL",
            ),
            "inClear-32K-Boost4Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost4Durx6Minutes",
                527,
                "ALL",
            ),
            "inClear-32K-Boost5Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost5Durx6Minutes",
                529,
                "ALL",
            ),
            "inClear-32K-Boost6Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost6Durx6Minutes",
                531,
                "ALL",
            ),
            "inClear-32K-Boost7Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost7Durx6Minutes",
                533,
                "ALL",
            ),
            "inClear-32K-Boost8Durx6Minutes": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-Boost8Durx6Minutes",
                535,
                "ALL",
            ),
            "inClear-32K-ValidRemoteFlow": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/All/inClear-32K-ValidRemoteFlow",
                537,
                "ALL",
            ),
        }



"""Structure accessor."""

from __future__ import annotations

import logging
import math
import struct
from typing import TYPE_CHECKING, Any

from geckolib.const import GeckoConstants

from .observable import Observable

if TYPE_CHECKING:
    from .async_spastruct import GeckoAsyncStructure

_LOGGER = logging.getLogger(__name__)


class GeckoStructAccessor(Observable):
    """
    Class to access the spa data structure.

    Uses the declaration in the specific modules.
    """

    @staticmethod
    def pack_data(length: int, data: Any) -> bytes:
        """Pack the data into bytes."""
        if isinstance(data, bytes):
            return data
        if length == 1:
            return struct.pack(">B", data)
        if length == 2:  # noqa: PLR2004
            return struct.pack(">H", data)
        raise OverflowError(length)

    def __init__(  # noqa: PLR0913
        self,
        struct_: GeckoAsyncStructure,
        path: str,
        pos: int,
        accessor_type: str,
        bitpos: int | None,
        items: str | list[str] | None,
        size: int | None,
        maxitems: int | None,
        rw: str | None,
    ) -> None:
        """Initialize the accessor class."""
        super().__init__()

        self.path: str = path
        self.tag: str = path.split("/")[-1]
        self.struct: GeckoAsyncStructure = struct_
        self.pos: int = pos
        self.accessor_type: str = accessor_type
        self.bitpos: int | None = bitpos
        self.items: list[str] | None = None
        self.maxitems: int | None = None
        self.direct_update: bool = False

        if bitpos is not None:
            self.bitmask = 1

        if items is not None:
            if isinstance(items, str):
                self.items = items.split("|")
            else:
                self.items = items

        self.length = 1
        self.format = ">B"

        if size is not None:
            self.length = size
            if self.length == 2:  # noqa: PLR2004
                self.format = ">H"

        if self.accessor_type in (
            GeckoConstants.SPA_PACK_STRUCT_WORD_TYPE,
            GeckoConstants.SPA_PACK_STRUCT_TIME_TYPE,
        ):
            self.length = 2
            self.format = ">H"

        if maxitems is not None:
            self.maxitems = int(maxitems)
            self.bitmask = (1 << math.ceil(math.log2(max(self.maxitems, 2)))) - 1

        self.read_write: str | None = rw

    def set_read_write(self, mode: str | None) -> None:
        """Set read/write state. Used by simulator."""
        self.read_write = mode

    def status_block_changed(self, offset: int, length: int, previous: bytes) -> None:
        """Status block has changed."""
        # Does the notified range intersect us, if not then we don't care!
        intersection_start = max(offset, self.pos)
        intersection_end = min(offset + length, self.pos + self.length)
        if intersection_end - intersection_start <= 0:
            return

        old_value = self._get_value(previous)
        new_value = self.value

        # No reason to notify if there is no change
        if new_value == old_value:
            return

        _LOGGER.info(
            "Status block changed value of %s from %s to %s",
            self.tag,
            old_value,
            new_value,
        )

        self._on_change(self, old_value, new_value)

    def _get_raw_value(self, status_block: bytes | None = None) -> Any:
        """
        Get a value from the pack structure.

        Either use the initialized declaration or using the optionally
        supplied status_block (used by change notification).
        """
        if status_block is None:
            status_block = self.struct.status_block
        data = struct.unpack(
            self.format, status_block[self.pos : self.pos + self.length]
        )[0]
        if self.bitpos is not None:
            data = (data >> self.bitpos) & self.bitmask
        return data

    def _get_value(self, status_block: bytes | None = None) -> Any:
        data = self._get_raw_value(status_block)
        if self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE:
            data = data == 1
        elif self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            try:
                assert self.items is not None  # noqa: S101
                data = self.items[data]
            except IndexError:
                data = "Unknown"
                _LOGGER.debug(
                    "Enum accessor %s out-of-range for %s. Value is %s",
                    self.tag,
                    self.items,
                    data,
                )
        elif self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_TIME_TYPE:
            data = f"{int(data / 256):02}:{data % 256:02}"
        return data

    @property
    def value(self) -> Any:
        """Get a value from the pack structure using the initialized declaration."""
        return self._get_value()

    @property
    def raw_value(self) -> Any:
        """
        Get a raw integer value from the pack structure.

        Uses the initialized declaration.
        """
        return self._get_raw_value()

    async def async_set_value(
        self, newvalue: Any, *, always_send: bool = False
    ) -> None:
        """Set a value in the pack structure using the initialized declaration."""
        if self.read_write is None:
            raise AttributeError(
                GeckoConstants.EXCEPTION_MESSAGE_NOT_WRITABLE.format(self.tag)
            )

        if self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            try:
                assert self.items is not None  # noqa: S101
                newvalue = self.items.index(f"{newvalue}")
            except ValueError:
                _LOGGER.exception(
                    "Can't set %s to %s, the list(%s) doesn't contain it",
                    self,
                    newvalue,
                    self.items,
                )
                raise
        elif self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_TIME_TYPE:
            bits = newvalue.split(":")
            newvalue = (int(bits[0]) * 256) + (int(bits[1]) % 256)
        elif (
            self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BYTE_TYPE
            and isinstance(newvalue, str)
        ) or (
            self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_WORD_TYPE
            and isinstance(newvalue, str)
        ):
            newvalue = int(newvalue)
        elif (
            self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE
            and isinstance(newvalue, str)
        ):
            newvalue = newvalue.lower() == "true"

        # If it is a bitpos, then mask it with the existing value
        existing = struct.unpack(
            self.format, self.struct.status_block[self.pos : self.pos + self.length]
        )[0]
        if self.bitpos is not None:
            newvalue = (existing & ~(self.bitmask << self.bitpos)) | (
                (newvalue & self.bitmask) << self.bitpos
            )

        if newvalue == existing and not always_send:
            return

        _LOGGER.debug(
            "Accessor %s @ %s, %s setting value to %s, existing value was %s. "
            "Length is %d",
            self.tag,
            self.pos,
            self.accessor_type,
            newvalue,
            existing,
            self.length,
        )

        if self.direct_update:
            self.struct.replace_status_block_segment(
                self.pos, GeckoStructAccessor.pack_data(self.length, newvalue)
            )
        else:
            # We can't handle this here, we must delegate via the structure
            await self.struct.async_set_value(self.pos, self.length, newvalue)

    def diag_info(self) -> str:
        """Get diagnostic data for this accessor."""
        info = f"{self.tag} = {self.value}\n"
        info += "-------------------------------\n"
        info += f"Path: {self.path}\n"
        info += f"Pos: {self.pos}\n"
        if self.bitpos:
            info += f"Bitpos: {self.bitpos}\n"
        info += f"Len: {self.length}\n"
        info += f"Type: {self.accessor_type}"
        if self.accessor_type == GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE:
            info += f" one of {self.items}"
        info += "\n"
        info += f"Raw Value: {self.raw_value}"
        info += f" ({self.struct.status_block[self.pos : self.pos + self.length]})\n"
        info += f"R/W: {self.read_write}"
        return info

    def __repr__(self) -> str:
        """Get string representation."""
        return f"{self.path!r}: {self.value!r}"

    def __eq__(self, other: GeckoStructAccessor) -> bool:
        """Equality comparator."""
        return self.value == other.value


class GeckoByteStructAccessor(GeckoStructAccessor):
    """Structure accessor for bytes."""

    def __init__(
        self, struct_: GeckoAsyncStructure, path: str, pos: int, rw: str | None
    ) -> None:
        """Initialize the byte accessor."""
        super().__init__(
            struct_,
            path,
            pos,
            GeckoConstants.SPA_PACK_STRUCT_BYTE_TYPE,
            None,
            None,
            None,
            None,
            rw,
        )


class GeckoWordStructAccessor(GeckoStructAccessor):
    """Structure accessor for words."""

    def __init__(
        self, struct_: GeckoAsyncStructure, path: str, pos: int, rw: str | None
    ) -> None:
        """Initialize the word accessor."""
        super().__init__(
            struct_,
            path,
            pos,
            GeckoConstants.SPA_PACK_STRUCT_WORD_TYPE,
            None,
            None,
            None,
            None,
            rw,
        )


class GeckoTimeStructAccessor(GeckoStructAccessor):
    """Structure accessor for times."""

    def __init__(
        self, struct_: GeckoAsyncStructure, path: str, pos: int, rw: str | None
    ) -> None:
        """Initialize the time accessor."""
        super().__init__(
            struct_,
            path,
            pos,
            GeckoConstants.SPA_PACK_STRUCT_TIME_TYPE,
            None,
            None,
            None,
            None,
            rw,
        )


class GeckoBoolStructAccessor(GeckoStructAccessor):
    """Structure accessor to bools."""

    def __init__(
        self,
        struct_: GeckoAsyncStructure,
        path: str,
        pos: int,
        bitpos: int | None,
        rw: str | None,
    ) -> None:
        """Initialie the bool accessor."""
        super().__init__(
            struct_,
            path,
            pos,
            GeckoConstants.SPA_PACK_STRUCT_BOOL_TYPE,
            bitpos,
            None,
            None,
            None,
            rw,
        )


class GeckoEnumStructAccessor(GeckoStructAccessor):
    """Structure accessopr for enumerations."""

    def __init__(  # noqa: PLR0913
        self,
        struct_: GeckoAsyncStructure,
        path: str,
        pos: int,
        bitpos: int | None,
        items: str | list[str] | None,
        size: int | None,
        maxitems: int | None,
        rw: str | None,
    ) -> None:
        """Initialize the enum accessor."""
        super().__init__(
            struct_,
            path,
            pos,
            GeckoConstants.SPA_PACK_STRUCT_ENUM_TYPE,
            bitpos,
            items,
            size,
            maxitems,
            rw,
        )


class GeckoTempStructAccessor(GeckoWordStructAccessor):
    """
    Handle Gecko temperatures.

    The spa handles all temperatures in tenths of a degree farenheit offset
    from freezing. This class handles that in one place.
    """

    def _get_value(self, status_block: bytes | None = None) -> float:
        """Get the temperature."""
        # Internally, temp is in farenheight tenths, offset from freezing point
        temp: float = super()._get_value(status_block)
        units = self.struct.accessors[GeckoConstants.KEY_TEMP_UNITS].value
        return temp / 18.0 if units == "C" else (temp + 320) / 10.0

    async def async_set_value(self, temp: float) -> None:
        """Set the temperature."""
        units = self.struct.accessors[GeckoConstants.KEY_TEMP_UNITS].value
        temp = float(temp) * 18.0 if units == "C" else float(temp) * 10.0 - 320
        await super().async_set_value(int(temp))



"""GeckoShell class."""

# ruff: noqa: T201

import datetime
import logging
import sys
import traceback
from time import sleep
from typing import Any, Self

from geckolib import VERSION, GeckoConstants, GeckoPump
from geckolib.async_spa_manager import GeckoAsyncSpaMan
from geckolib.automation.base import GeckoAutomationBase, GeckoAutomationFacadeBase
from geckolib.config import release_config_change_waiters
from geckolib.driver.accessor import GeckoStructAccessor
from geckolib.spa_events import GeckoSpaEvent

from .shared_command import GeckoCmd

_LOGGER = logging.getLogger(__name__)


SHELL_UUID = "02ac6d28-42d0-41e3-ad22-274d0aa491da"


class GeckoShell(GeckoCmd, GeckoAsyncSpaMan):
    """Client shell to explore spa capabilities."""

    BANNER = """

        <Disclaimer>
        ----------------------------- USE AT YOUR OWN RISK -----------------------------

        This code will allow you to make changes to your spa configuration that is
        outside of what the app, top panel and side panel settings allow. I've not
        tested every setting and it might be that you prevent your spa pack from
        operating as it used to do.

        I strongly suggest dumping the configuration values with the "config" command
        and recording them somewhere safe.

        </Disclaimer>

    """

    def __init__(self) -> None:
        """Initialize the GeckoShell class."""
        self.spas = None
        GeckoAsyncSpaMan.__init__(self, SHELL_UUID)
        GeckoCmd.__init__(self, self)

        # Fill the watercare mode strings in.
        assert self.do_watercare.__doc__ is not None  # noqa: S101
        self.do_watercare.__func__.__doc__ = self.do_watercare.__doc__.format(
            GeckoConstants.WATERCARE_MODE_STRING
        )

        self.intro = "Welcome to the Gecko shell. Type help or ? to list commands.\n"
        self.prompt = "(Gecko) "

    async def __aenter__(self) -> Self:
        """Support 'with'."""
        _LOGGER.info("Async enter called on %s", self.__class__.__name__)
        await GeckoAsyncSpaMan.__aenter__(self)
        return self

    def _on_accessor_changed(
        self, accessor: GeckoStructAccessor, old_value: Any, new_value: Any
    ) -> None:
        print(f"accessor {accessor} changed from {old_value} to {new_value}")

    async def handle_event(self, event: GeckoSpaEvent, **_kwargs: object) -> None:  # noqa: PLR0912
        """Handle the event."""
        _LOGGER.debug("Handle event %s", event)
        if event == GeckoSpaEvent.CLIENT_FACADE_IS_READY:
            if self._spa is not None:
                self.structure = self._spa.struct
                for accessor in self.structure.accessors.values():
                    accessor.watch(self._on_accessor_changed)

        elif event == GeckoSpaEvent.CLIENT_FACADE_TEARDOWN:
            self.structure = None

        elif event == GeckoSpaEvent.LOCATING_STARTED:
            print("Starting discovery process...", end="", flush=True)

        elif event == GeckoSpaEvent.LOCATING_FINISHED:
            self.spas = self._spa_descriptors
            number_of_spas = len(self.spas)
            print(f"Found {number_of_spas} spas", flush=True)
            if number_of_spas == 0:
                _LOGGER.warning(
                    "Try using the iOS or Android app to confirm they are "
                    "functioning correctly"
                )
                sys.exit(1)
            if number_of_spas == 1:
                self.do_manage("1")
            else:
                print(self.prompt, end="", flush=True)

        elif event == GeckoSpaEvent.CONNECTION_FINISHED:
            assert self.facade is not None  # noqa: S101
            print(f"connected to {self.facade.name}!", flush=True)
            self.prompt = f"{self.facade.name}$ "

            # Build list of spa commands
            for device in self.facade.all_user_devices:
                if isinstance(device, GeckoPump):
                    func_name = f"do_{device.key}"

                    async def async_pump_command(
                        self: Self, arg: str, device: GeckoPump = device
                    ) -> None:
                        await self.pump_command(arg, device)

                    setattr(
                        GeckoShell,
                        func_name,
                        async_pump_command,
                    )
                    func_ptr = getattr(GeckoShell, func_name)
                    func_ptr.__doc__ = (
                        f"Set pump {device.name} mode: {device.key} <OFF|LO|HI>"
                    )
                else:
                    func_name = f"do_{device.key}"

                    async def async_device_command(
                        self: Self, arg: str, device: GeckoAutomationBase = device
                    ) -> None:
                        await self.device_command(arg, device)

                    setattr(
                        GeckoShell,
                        func_name,
                        async_device_command,
                    )
                    func_ptr = getattr(GeckoShell, func_name)
                    func_ptr.__doc__ = (
                        f"Turn device {device.name} ON or OFF: {device.key} <ON|OFF>"
                    )

            self.do_state("")
            print(self.prompt, end="", flush=True)

    def do_discover(self, arg: str) -> None:
        """
        Discover all the in.touch2 devices on your network.

        usage: discover [<ip address>].
        """
        if self.spas is not None:
            return
        self._spa_address = arg

    def do_list(self, _arg: str) -> None:
        """List the spas that are available to manage : list."""
        for idx, spa in enumerate(self.spas):
            print(f"{idx + 1}. {spa.name} ({spa.ipaddress})")

    def do_manage(self, arg: str) -> None:
        """Manage a named or numbered spa : manage 1."""
        spa_to_manage = int(arg)
        spa_descriptor = self.spas[spa_to_manage - 1]
        print(
            (
                f"Connecting to spa `{spa_descriptor.name}`"
                f" at {spa_descriptor.ipaddress} ... "
            ),
            end="",
            flush=True,
        )
        _LOGGER.debug(spa_descriptor)
        self._spa_identifier = spa_descriptor.identifier_as_string
        self._state_change.set()

    async def device_command(self, arg: str, device: GeckoAutomationFacadeBase) -> None:
        """Turn a device on or off."""
        print(f"Turn device {device.name} {arg}")
        if arg.lower() == "on":
            await device.async_turn_on()
        else:
            await device.async_turn_off()

    async def pump_command(self, arg: str, device: GeckoPump) -> None:
        """Set a pump mode <mode>."""
        print(f"Set pump {device.name} {arg}")
        try:
            await device.async_set_mode(arg)
        except Exception:  # noqa: BLE001
            traceback.print_exc()

    async def do_key(self, arg: str) -> None:
        """
        Press keypad button.

        Usage: key <keynum>. If keynum missing, then defaults to 1.
        """
        keypad = GeckoConstants.KEYPAD_PUMP_1
        if arg:
            try:
                keypad = int(arg)
            except ValueError:
                print("Numbers only please")
                print(self.prompt, end="", flush=True)
                return
        await self.facade.spa.async_press(keypad)

    def do_state(self, _arg: str) -> None:
        """Show the state of the managed spa : state."""
        if self.facade is None:
            print("Must be connected to a spa")
            return
        print(self.facade.water_heater)
        for pump in self.facade.pumps:
            print(pump)
        for blower in self.facade.blowers:
            print(blower)
        for light in self.facade.lights:
            print(light)
        if self.facade.reminders_manager.is_available:
            for reminder in self.facade.reminders_manager.reminders:
                print(reminder)
        if self.facade.water_care.is_available:
            print(self.facade.water_care)
        for sensor in [
            *self.facade.sensors,
            *self.facade.binary_sensors,
        ]:
            print(sensor)
        if self.facade.eco_mode is not None:
            print(self.facade.eco_mode)
        print(self.facade.error_sensor)
        if self.facade.ingrid.is_available:
            print(self.facade.ingrid)
        if self.facade.heatpump.is_available:
            print(self.facade.heatpump)
        if self.facade.inmix.is_available:
            print(self.facade.inmix)

    def monitor_get_states(self) -> list[str]:
        """Get the monitor states as a string list."""
        states = [
            self.facade.water_heater,
            *self.facade.pumps,
            *self.facade.blowers,
            *self.facade.lights,
            # *self.facade.reminders,
            self.facade.water_care,
            *self.facade.sensors,
            *self.facade.binary_sensors,
            self.facade.eco_mode,
        ]
        return [f"{state.monitor}" for state in states]

    def monitor_compare_states(self, states: list[str]) -> bool:
        """Compare the current monitor states."""
        local_state = self.monitor_get_states()
        return local_state != states

    def monitor_print_states(self, states: list[str]) -> None:
        """Print the monitor states out."""
        print(f"{datetime.datetime.now(tz=datetime.UTC)} : {' '.join(states)}")

    def do_monitor(self, _arg: str) -> None:
        """
        Monitor the state of the managed spa.

        Output a new line for each change

        usage: monitor.
        """
        if self.facade is None:
            print("Must be connected to a spa")
            return

        print("Monitoring spa ... CTRL+C to stop")
        current_state = []
        while True:
            try:
                if self.monitor_compare_states(current_state):
                    current_state = self.monitor_get_states()
                    self.monitor_print_states(current_state)
                sleep(1)
            except KeyboardInterrupt:
                print()
                print("Monitor stopped")
                break

    @property
    def version_strings(self) -> list[str]:
        """Get the version strings for the spa."""
        return [
            f"geckolib version {VERSION}",
            f"SpaPackStruct.xml revision {self.facade.spa.revision}",
            f"intouch version EN {self.facade.spa.intouch_version_en}",
            f"intouch version CO {self.facade.spa.intouch_version_co}",
            f"Spa pack {self.facade.spa.pack} {self.facade.spa.version}",
            f"Low level configuration # {self.facade.spa.config_number}",
            f"Config version {self.facade.spa.config_version}",
            f"Log version {self.facade.spa.log_version}",
            f"Pack type {self.facade.spa.pack_type}",
        ]

    def do_version(self, _arg: str) -> None:
        """Show the version information : version."""
        for version_str in self.version_strings:
            print(version_str)

    def do_about(self, _arg: str) -> None:
        """Display information about this client program and support library : about."""
        print()
        print(
            "GeckoShell: A python program using GeckoLib library to drive Gecko enabled"
            " devices with in.touch2 communication modules"
        )
        print(f"Library version v{VERSION}")

    async def do_watercare(self, arg: str) -> None:
        """Set the active watercare mode to one of {0} : WATERCARE <mode>."""
        try:
            await self.facade.water_care.async_set_mode(arg)
        except Exception:  # pylint: disable=broad-except
            _LOGGER.exception("Exception setting watercare to '%s'", arg)

    async def do_setpoint(self, arg: str) -> None:
        """Set the spa setpoint temperature : setpoint <temp>."""
        await self.facade.water_heater.async_set_target_temperature(float(arg))

    def do_eco(self, arg: str) -> None:
        """Set the spa eco mode : eco on|off."""
        assert self.facade is not None  # noqa: S101
        assert self.facade.eco_mode is not None  # noqa: S101
        if arg.lower() == "off":
            self.facade.eco_mode.turn_off()
        else:
            self.facade.eco_mode.turn_on()

    def get_snapshot_data(self) -> dict:
        """Proxy for the spa."""
        assert self._facade is not None  # noqa: S101
        return self._facade.spa.get_snapshot_data()

    def do_refresh(self, _args: str) -> None:
        """Refresh the spa pack data."""
        release_config_change_waiters()



"""Gecko Water Heaters."""

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from geckolib.automation.power import GeckoPower
from geckolib.const import GeckoConstants

from .sensors import GeckoBinarySensor, GeckoSensor

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoWaterHeaterAbstract(GeckoPower):
    """Abstract base for water heaters."""

    TEMP_CELCIUS = "°C"
    TEMP_FARENHEIGHT = "°F"

    @property
    @abstractmethod
    def current_operation(self) -> str:
        """Get the current operation."""

    @property
    @abstractmethod
    def temperature_unit(self) -> str:
        """Get the temperature unit."""

    @property
    @abstractmethod
    def current_temperature(self) -> float:
        """Get the current temperature."""

    @property
    @abstractmethod
    def target_temperature(self) -> float:
        """Get the target temperature."""

    @property
    @abstractmethod
    def min_temp(self) -> float:
        """Get the min temp."""

    @property
    @abstractmethod
    def max_temp(self) -> float:
        """Get the max temp."""


class GeckoWaterHeater(GeckoWaterHeaterAbstract):
    """Water Heater object based on Home Assistant Entity Type Climate."""

    MIN_TEMP_C = 15
    MAX_TEMP_C = 40
    MIN_TEMP_F = 59
    MAX_TEMP_F = 104

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the water heater class."""
        super().__init__(facade, "Heater", "HEAT")

        self._current_temperature_sensor = None
        self._target_temperature_sensor = None
        self._real_setpoint_sensor = None
        self._temperature_unit_accessor = None
        self._heating_action_sensor = None
        self._cooling_action_sensor = None
        self._min_temp_sensor = None
        self._max_temp_sensor = None
        self._temp_not_valid_sensor = None

        # Attempt to locate the various items needed from the spa accessors
        if GeckoConstants.KEY_TEMP_UNITS in self._spa.accessors:
            self._temperature_unit_accessor = self._spa.accessors[
                GeckoConstants.KEY_TEMP_UNITS
            ]
        if GeckoConstants.KEY_SETPOINT_G in self._spa.accessors:
            self._target_temperature_sensor = GeckoSensor(
                facade,
                "Target Temperature",
                self._spa.accessors[GeckoConstants.KEY_SETPOINT_G],
                self._temperature_unit_accessor,
            )
            self.set_availability(is_available=True)
        if GeckoConstants.KEY_DISPLAYED_TEMP_G in self._spa.accessors:
            self._current_temperature_sensor = GeckoSensor(
                facade,
                "Current Temperature",
                self._spa.accessors[GeckoConstants.KEY_DISPLAYED_TEMP_G],
                self._temperature_unit_accessor,
            )
            self.set_availability(is_available=True)
        if GeckoConstants.KEY_REAL_SETPOINT_G in self._spa.accessors:
            self._real_setpoint_sensor = GeckoSensor(
                facade,
                "Real Target Temperature",
                self._spa.accessors[GeckoConstants.KEY_REAL_SETPOINT_G],
                self._temperature_unit_accessor,
            )
            self.set_availability(is_available=True)

        self._heating_action_sensor = self._cooling_action_sensor = None
        if GeckoConstants.KEY_HEATING in self._spa.accessors:
            self._heating_action_sensor = GeckoBinarySensor(
                self.facade, "Heating", self._spa.accessors[GeckoConstants.KEY_HEATING]
            )
        if GeckoConstants.KEY_COOLINGDOWN in self._spa.accessors:
            self._cooling_action_sensor = GeckoBinarySensor(
                self.facade,
                "Cooling",
                self._spa.accessors[GeckoConstants.KEY_COOLINGDOWN],
            )
        if GeckoConstants.KEY_MIN_SETPOINT_G in self._spa.accessors:
            self._min_temp_sensor = GeckoSensor(
                self.facade,
                "Min Temperature",
                self._spa.accessors[GeckoConstants.KEY_MIN_SETPOINT_G],
                self._temperature_unit_accessor,
            )
        if GeckoConstants.KEY_MAX_SETPOINT_G in self._spa.accessors:
            self._max_temp_sensor = GeckoSensor(
                self.facade,
                "Max Temperature",
                self._spa.accessors[GeckoConstants.KEY_MAX_SETPOINT_G],
                self._temperature_unit_accessor,
            )
        if GeckoConstants.KEY_TEMP_NOT_VALID in self._spa.accessors:
            self._temp_not_valid_sensor = GeckoSensor(
                self.facade,
                "Temp Not Valid",
                self._spa.accessors[GeckoConstants.KEY_TEMP_NOT_VALID],
            )

        # Setup change observers
        for sensor in [
            self._current_temperature_sensor,
            self._target_temperature_sensor,
            self._real_setpoint_sensor,
            self._temperature_unit_accessor,
            self._heating_action_sensor,
            self._cooling_action_sensor,
            self._min_temp_sensor,
            self._max_temp_sensor,
            self._temp_not_valid_sensor,
        ]:
            if sensor is not None:
                sensor.watch(self._on_change)

        self._on_change(None, None, None)

    @property
    def target_temperature_sensor(self) -> GeckoSensor:
        """Get the target temperature sensor object."""
        return self._target_temperature_sensor

    @property
    def target_temperature(self) -> float:
        """Get the target temperature of the water."""
        return self._target_temperature_sensor.state

    async def async_set_target_temperature(self, new_temperature: float) -> None:
        """Set the target temperature of the water."""
        await self._target_temperature_sensor.accessor.async_set_value(new_temperature)

    @property
    def real_target_temperature(self) -> float:
        """Get the real target temperature (takes eco mode into account)."""
        if self._real_setpoint_sensor is not None:
            return self._real_setpoint_sensor.state
        return self.target_temperature

    @property
    def real_target_temperature_sensor(self) -> GeckoSensor:
        """Get the real target temperature sensor (takes eco mode into account)."""
        return self._real_setpoint_sensor

    @property
    def min_temp(self) -> float:
        """Get the minimum temperature of the water heater."""
        if self._min_temp_sensor is not None:
            return self._min_temp_sensor.state
        return (
            self.MIN_TEMP_C
            if self._temperature_unit_accessor.value == "C"
            else self.MIN_TEMP_F
        )

    @property
    def max_temp(self) -> float:
        """Get the maximum temperature of the water heater."""
        if self._max_temp_sensor is not None:
            return self._max_temp_sensor.state
        return (
            self.MAX_TEMP_C
            if self._temperature_unit_accessor.value == "C"
            else self.MAX_TEMP_F
        )

    @property
    def current_temperature(self) -> float:
        """Get the current temperature of the water."""
        return self._current_temperature_sensor.state

    @property
    def temperature_unit(self) -> str:
        """Get the temperature units for the water heater."""
        if self._temperature_unit_accessor.value == "C":
            return self.TEMP_CELCIUS
        return self.TEMP_FARENHEIGHT

    @property
    def current_temperature_sensor(self) -> GeckoSensor:
        """Get the current temperature sensor object."""
        return self._current_temperature_sensor

    async def async_set_temperature_unit(self, new_unit: str) -> None:
        """Set the temperature units for the water heater."""
        if new_unit in (self.TEMP_FARENHEIGHT, "f", "F"):
            await self._temperature_unit_accessor.async_set_value("F")
        else:
            await self._temperature_unit_accessor.async_set_value("C")

    @property
    def current_operation(self) -> str:  # noqa: PLR0911
        """Return the current operation of the water heater."""
        # If we have both sensors, then these are the arbiters
        if (
            self._heating_action_sensor is not None
            and self._cooling_action_sensor is not None
        ):
            if self._heating_action_sensor.is_on:
                return GeckoConstants.WATER_HEATER_HEATING
            if self._cooling_action_sensor.is_on:
                return GeckoConstants.WATER_HEATER_COOLING
            return GeckoConstants.WATER_HEATER_IDLE

        # Otherwise, we take what we can get
        if (
            self._heating_action_sensor is not None
            and self._heating_action_sensor.is_on
        ):
            return GeckoConstants.WATER_HEATER_HEATING
        if (
            self._cooling_action_sensor is not None
            and self._cooling_action_sensor.is_on
        ):
            return GeckoConstants.WATER_HEATER_COOLING

        # Finally, it's down to the actual temperatures
        if self.current_temperature < self.real_target_temperature:
            return GeckoConstants.WATER_HEATER_HEATING
        if self.current_temperature > self.real_target_temperature:
            return GeckoConstants.WATER_HEATER_COOLING
        return GeckoConstants.WATER_HEATER_IDLE

    def format_temperature(self, temperature: float) -> str:
        """Format a temperature value to a printable string."""
        return f"{temperature:.1f}{self.temperature_unit}"

    def __str__(self) -> str:
        """Stringize the class."""
        if self.is_available:
            return (
                f"{self.name}: Temperature "
                f"{self.format_temperature(self.current_temperature)}, SetPoint "
                f"{self.format_temperature(self.target_temperature)}, Real SetPoint "
                f"{self.format_temperature(self.real_target_temperature)}, Operation "
                f"{self.current_operation}"
            )
        return f"{self.name}: Not present"

    def _on_change(
        self, sender: Any = None, old_value: Any = None, new_value: Any = None
    ) -> None:
        if self._temp_not_valid_sensor is not None:
            self.set_availability(is_available=not self._temp_not_valid_sensor.state)
        return super()._on_change(sender, old_value, new_value)

    @property
    def monitor(self) -> str:
        """Get monitor string."""
        return (
            f"HTR: {self.format_temperature(self.current_temperature)}"
            f" SET: {self.format_temperature(self.real_target_temperature)}"
        )



"""Mr.Steam class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.automation.heater import GeckoWaterHeaterAbstract
from geckolib.automation.number import GeckoNumber
from geckolib.automation.sensors import GeckoSensor
from geckolib.automation.switch import GeckoSwitch
from geckolib.const import GeckoConstants

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class MrSteam(GeckoWaterHeaterAbstract):
    """Mr.Steam support class."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the MrSteam class."""
        super().__init__(facade, "Steamer", "MRSTEAM")
        if not facade.spa.struct.is_mr_steam:
            return

        self.set_availability(is_available=True)

        self.onoff = GeckoSwitch(
            facade,
            GeckoConstants.KEY_MRSTEAM_USER_MODE,
            (
                "On/Off",
                GeckoConstants.KEY_MRSTEAM_USER_MODE,
                GeckoConstants.DEVICE_CLASS_SWITCH,
            ),
        )
        self.onoff.watch(self._on_change)

        self.aroma = GeckoSwitch(
            facade,
            GeckoConstants.KEY_MRSTEAM_USER_AROMA,
            (
                "Aroma",
                GeckoConstants.KEY_MRSTEAM_USER_AROMA,
                GeckoConstants.DEVICE_CLASS_SWITCH,
            ),
        )
        self.aroma.watch(self._on_change)

        self.chroma = GeckoSwitch(
            facade,
            GeckoConstants.KEY_MRSTEAM_USER_CHROMA,
            (
                "Chroma",
                GeckoConstants.KEY_MRSTEAM_USER_CHROMA,
                GeckoConstants.DEVICE_CLASS_SWITCH,
            ),
        )
        self.chroma.watch(self._on_change)

        self.setpoint_accessor = self._spa.accessors[
            GeckoConstants.KEY_MRSTEAM_USER_SETPOINT_G
        ]
        self.setpoint_accessor.watch(self._on_change)
        self.temp_unt_accesor = self._spa.accessors[GeckoConstants.KEY_TEMP_UNITS]
        self.min_temp_accessor = self._spa.accessors[GeckoConstants.KEY_MIN_SETPOINT_G]
        self.max_temp_accessor = self._spa.accessors[GeckoConstants.KEY_MAX_SETPOINT_G]

        self.user_runtime = GeckoNumber(
            facade, "Runtime", GeckoConstants.KEY_MRSTEAM_USER_RUNTIME, "min"
        )
        self.user_runtime.watch(self._on_change)
        if GeckoConstants.KEY_MRSTEAM_MAX_RUNTIME in self._spa.accessors:
            max_runtime = self._spa.accessors[GeckoConstants.KEY_MRSTEAM_MAX_RUNTIME]
            self.user_runtime.native_max_value = max_runtime.value

        self.remaining_runtime = GeckoSensor(
            facade,
            "Remaining Runtime",
            self._spa.accessors[GeckoConstants.KEY_MRSTEAM_REMAINING_RUNTIME],
            "min",
        )
        self.remaining_runtime.watch(self._on_change)

    @property
    def switches(self) -> list[GeckoSwitch]:
        """Get a list of all the switches."""
        if self.is_available:
            return [self.onoff, self.aroma, self.chroma]
        return []

    @property
    def current_operation(self) -> str:
        """Get the current operation."""
        if not self.onoff.is_on:
            return "Turned off"
        return f"Running, {self.remaining_runtime.state}m remaining"

    @property
    def temperature_unit(self) -> str:
        """Get the temperature unit."""
        if self.temp_unt_accesor.value == "C":
            return self.TEMP_CELCIUS
        return self.TEMP_FARENHEIGHT

    @property
    def current_temperature(self) -> float:
        """Get the current temperature."""
        return self.setpoint_accessor.value

    @property
    def target_temperature(self) -> float:
        """Get the target temperature."""
        return self.setpoint_accessor.value

    @property
    def min_temp(self) -> float:
        """Get the min temp."""
        return self.min_temp_accessor.value

    @property
    def max_temp(self) -> float:
        """Get the max temp."""
        return self.max_temp_accessor.value

    async def async_set_target_temperature(self, new_temperature: float) -> None:
        """Set the target temperature of the water."""
        await self.setpoint_accessor.async_set_value(new_temperature)



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InXM v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 2

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return [
            "Out1A",
            "Out1B",
            "Out2A",
            "Out2B",
            "Out3A",
            "OutHeater",
            "Out4A",
            "Out5A",
            "Out6A",
            "Out7A",
            "OutLi",
            "OutIO2",
            "OutIO4",
            "LightActivationOrder",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "ConfigNumber": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/ConfigNumber", 1, "ALL"
            ),
            "Out1A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1A",
                200,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out1B": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1B",
                201,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2A",
                202,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2B": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2B",
                203,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out3A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out3A",
                204,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "OutHeater": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/OutHeater",
                209,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "HTR"],
                None,
                None,
                "ALL",
            ),
            "Out1ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1ACurrent", 220, "ALL"
            ),
            "Out1BCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1BCurrent", 221, "ALL"
            ),
            "Out2ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2ACurrent", 222, "ALL"
            ),
            "Out2BCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2BCurrent", 223, "ALL"
            ),
            "Out3ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out3ACurrent", 224, "ALL"
            ),
            "Out4A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out4A",
                205,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out5A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out5A",
                206,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out6A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out6A",
                207,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out7A": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out7A",
                208,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "Out4ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out4ACurrent", 225, "ALL"
            ),
            "Out5ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out5ACurrent", 226, "ALL"
            ),
            "Out6ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out6ACurrent", 227, "ALL"
            ),
            "Out7ACurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out7ACurrent", 228, "ALL"
            ),
            "DirectCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/DirectCurrent", 123, "ALL"
            ),
            "OutLi": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutLi",
                210,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "LI"],
                None,
                None,
                "ALL",
            ),
            "OutIO2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutIO2",
                211,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "OutIO4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutIO4",
                212,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "LI",
                    "FAN",
                    "FbMtr",
                    "FbLi",
                    "FullOn",
                    "TvLift",
                    "SpLift",
                    "SANI",
                    "ONZEN",
                    "VALVE",
                ],
                None,
                None,
                "ALL",
            ),
            "LightInts": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LVOutputConfig/LightInts", 84, "ALL"
            ),
            "LightActivationOrder": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/LightActivationOrder",
                17,
                0,
                ["Ascendant", "Descendant"],
                None,
                2,
                "ALL",
            ),
            "LightTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/LightTimeOut", 18, "ALL"
            ),
            "FiberTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/FiberTimeOut", 19, "ALL"
            ),
            "LightResetTimeOut": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/TimeOut/LightResetTimeOut",
                17,
                1,
                ["OnStart", "OnChange"],
                None,
                2,
                "ALL",
            ),
            "PumpTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOut", 20, "ALL"
            ),
            "PumpResetTimeOut": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/TimeOut/PumpResetTimeOut",
                17,
                2,
                ["OnStart", "OnChange"],
                None,
                2,
                "ALL",
            ),
            "PumpTimeOutShare": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOutShare", 17, 3, "ALL"
            ),
            "QuietTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/QuietTimeOut", 69, "ALL"
            ),
            "FiltInterface": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltInterface",
                35,
                None,
                ["StartDurFre", "NotUsed1", "NotUsed2", "PurgeOnly"],
                None,
                None,
                "ALL",
            ),
            "CleanP1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CleanP1", 33, 1, "ALL"
            ),
            "CleanCP": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CleanCP", 33, 2, "ALL"
            ),
            "FiltOTOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/FiltOTOption", 34, 0, "ALL"
            ),
            "CPAlwaysON": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/CPAlwaysON", 33, 3, "ALL"
            ),
            "FiltCPOTOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/FilterConfig/FiltCPOTOption", 34, 1, "ALL"
            ),
            "FiltSuspendBy": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltSuspendBy",
                34,
                2,
                ["AnyUD", "PumpUD"],
                None,
                2,
                "ALL",
            ),
            "FiltCPOTLimitOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltCPOTLimitOption",
                34,
                3,
                ["AlwaysEnabled", "LimitedIfSPBelow95F"],
                None,
                2,
                "ALL",
            ),
            "PurgeSpeed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/PurgeSpeed",
                33,
                0,
                ["Lo", "Hi"],
                None,
                2,
                "ALL",
            ),
            "FiltMinDur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/AdvanceFilterConfig/FiltMinDur", 57, "ALL"
            ),
            "FiltOTActTemp": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOTActTemp",
                55,
                "ALL",
            ),
            "FiltOTDeadband": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOTDeadband",
                56,
                "ALL",
            ),
            "FiltSuspendTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltSuspendTime",
                62,
                "ALL",
            ),
            "FiltCPOTActTemp": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTActTemp",
                58,
                "ALL",
            ),
            "FiltCPOTDeadband": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTDeadband",
                59,
                "ALL",
            ),
            "FiltCPMinOnTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPMinOnTime",
                60,
                "ALL",
            ),
            "FiltCPOTMaxDur": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltCPOTMaxDur",
                61,
                "ALL",
            ),
            "O3Pump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3/O3Pump",
                63,
                0,
                ["NONE", "P1", "CP", ""],
                None,
                4,
                "ALL",
            ),
            "O3FollowPump": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/O3/O3FollowPump", 63, 2, "ALL"
            ),
            "O3DuringFilt": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/O3/O3DuringFilt", 63, 3, "ALL"
            ),
            "O3Toggling": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/O3/O3Toggling", 63, 4, "ALL"
            ),
            "O3SuspendBy": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3/O3SuspendBy",
                63,
                6,
                ["AnyUD", "PumpUD"],
                None,
                None,
                "ALL",
            ),
            "O3FilterDelay": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3/O3FilterDelay", 64, "ALL"
            ),
            "O3SuspendTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3/O3SuspendTime", 65, "ALL"
            ),
            "O3TogglePeriod": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3/O3TogglePeriod", 66, "ALL"
            ),
            "FiltFreq": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltFreq",
                36,
                "ALL",
            ),
            "FiltStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltStart",
                37,
                "ALL",
            ),
            "FiltDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltDur", 39, "ALL"
            ),
            "EconStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStart",
                71,
                "ALL",
            ),
            "EconStop": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStop",
                73,
                "ALL",
            ),
            "EconControlableManually": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconControlableManually",
                68,
                3,
                "ALL",
            ),
            "EconProgStatus": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconProgStatus",
                68,
                4,
                ["Disabled", "NotActive", "NA", "Active"],
                None,
                4,
                "ALL",
            ),
            "BlowerKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/BlowerKeyOption",
                17,
                4,
                ["LastPumpKey", "NextFreePumpKey", "AnyFreePump"],
                None,
                4,
                "ALL",
            ),
            "QuickOnOffKeyEnable": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/QuickOnOffKeyEnable",
                17,
                6,
                "ALL",
            ),
            "K600Msg1NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg1NbWeek", 75, "ALL"
            ),
            "K600Msg2NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg2NbWeek", 76, "ALL"
            ),
            "K600Msg3NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg3NbWeek", 77, "ALL"
            ),
            "K600Msg4NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg4NbWeek", 78, "ALL"
            ),
            "K600Msg5NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg5NbWeek", 79, "ALL"
            ),
            "K600Msg6NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg6NbWeek", 80, "ALL"
            ),
            "K600Msg7NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg7NbWeek", 81, "ALL"
            ),
            "K600Msg8NbWeek": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/K600Reminders/K600Msg8NbWeek", 82, "ALL"
            ),
            "SWMPurgeSpeed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SWM/SWMPurgeSpeed",
                67,
                0,
                ["Low", "High"],
                None,
                2,
                "ALL",
            ),
            "SWMPurgeBlower": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/SWM/SWMPurgeBlower", 67, 1, "ALL"
            ),
            "LineFrequency": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/LineFrequency",
                83,
                4,
                ["60HZ", "50HZ"],
                None,
                2,
                "ALL",
            ),
            "LineVoltage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/LineVoltage",
                83,
                5,
                ["240Vac", "230VAC"],
                None,
                2,
                "ALL",
            ),
            "NbPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/NbPhases", 85, None
            ),
            "BrNbSettings": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BrNbSettings", 86, None
            ),
            "BrSetIndex": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BrSetIndex", 87, None
            ),
            "BreakerSet1": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet1", 88, "ALL"
            ),
            "BreakerSet2": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet2", 89, "ALL"
            ),
            "BreakerSet3": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet3", 90, "ALL"
            ),
            "BreakerSet4": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet4", 91, "ALL"
            ),
            "BreakerSet5": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/BreakerSet5", 92, "ALL"
            ),
            "BreakerMenu": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/BreakerMenu",
                83,
                7,
                ["Available", "NotAvailable"],
                None,
                2,
                "ALL",
            ),
            "FuseMapping": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/FuseMapping",
                83,
                1,
                [
                    "UL",
                    "CE_Jmp1_3_Installed",
                    "CE_Jmp1_4_Installed",
                    "CE_Jmp2_3_Installed",
                    "CE_Jmp2_4_Installed",
                    "",
                    "",
                    "",
                ],
                None,
                8,
                "ALL",
            ),
            "QuietActionOnHeater": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/QuietActionOnHeater",
                23,
                3,
                ["NoAction", "ForceMaxPower", "KeepOff", ""],
                None,
                4,
                "ALL",
            ),
            "UDActionOnQuiet": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/UDActionOnQuiet",
                68,
                1,
                [
                    "IgnoreUD",
                    "ExitQuietAndRestoreAllUD",
                    "NA",
                    "ExitQuietAndCancelOtherUD",
                ],
                None,
                4,
                "ALL",
            ),
            "QuietActionOnUD": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Quiet/QuietActionOnUD",
                68,
                0,
                ["Cancel", "Pause"],
                None,
                2,
                "ALL",
            ),
            "FanPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Fan/FanPump",
                24,
                1,
                ["None", "P1", "CP"],
                None,
                4,
                "ALL",
            ),
            "FanAsCoolingDevice": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanAsCoolingDevice", 24, 3, "ALL"
            ),
            "FanActiveOnAmbiantOT": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveOnAmbiantOT", 23, 5, "ALL"
            ),
            "FanActiveOnPumpRun": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveOnPumpRun", 23, 6, "ALL"
            ),
            "FanForceOFFOption": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/Fan/FanForceOFFOption", 23, 7, "ALL"
            ),
            "FanSetAOTTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanSetAOTTempAdc", 27, "ALL"
            ),
            "FanClrAOTTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanClrAOTTempAdc", 28, "ALL"
            ),
            "FanActiveDelayAfterPumpRun": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanActiveDelayAfterPumpRun", 29, "ALL"
            ),
            "FanSetForceOffTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanSetForceOffTempAdc", 30, "ALL"
            ),
            "FanClrForceOffTempAdc": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanClrForceOffTempAdc", 31, "ALL"
            ),
            "FanForceOffInhibitDelay": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/Fan/FanForceOffInhibitDelay", 32, "ALL"
            ),
            "HeaterPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeaterPump",
                23,
                None,
                ["", "CP", "P1"],
                None,
                4,
                "ALL",
            ),
            "HeatRestriction": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeatRestriction",
                83,
                6,
                ["No", "FullPowerOnly"],
                None,
                2,
                "ALL",
            ),
            "ExtProbeEnable": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/ExtProbeEnable",
                24,
                4,
                "ALL",
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/TempUnits",
                23,
                2,
                ["C", "F"],
                None,
                2,
                "ALL",
            ),
            "SetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/SetpointG",
                25,
                "ALL",
            ),
            "EconBelowSetpoint": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/EconBelowSetpoint",
                70,
                "ALL",
            ),
            "AmbiantOTEnable": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/AmbiantOTEnable", 24, 0, "ALL"
            ),
            "TimeFormat": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/TimeFormat",
                83,
                0,
                ["AmPm", "24h"],
                None,
                2,
                "ALL",
            ),
            "Aux1AsTVLifter": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/Aux1AsTVLifter", 21, 0, "ALL"
            ),
            "Aux2AsSpeakerLifter": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/Aux2AsSpeakerLifter", 21, 1, "ALL"
            ),
            "DJS_No1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/DJS_No1",
                21,
                2,
                ["NA", "OnP1", "OnP2", "OnP3"],
                None,
                4,
                "ALL",
            ),
            "DJS_No2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/DJS_No2",
                21,
                4,
                ["NA", "OnP1", "OnP2", "OnP3"],
                None,
                4,
                "ALL",
            ),
        }



"""Gecko APING handlers."""

from __future__ import annotations

import logging
import struct
from typing import Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

PING_VERB = b"APING"

_LOGGER = logging.getLogger(__name__)


class GeckoPingProtocolHandler(GeckoPacketProtocolHandler):
    """Ping handler class."""

    @staticmethod
    def request(**kwargs: Any) -> GeckoPingProtocolHandler:
        """Build a PING request class."""
        return GeckoPingProtocolHandler(
            content=PING_VERB, timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS, **kwargs
        )

    @staticmethod
    def response(**kwargs: Any) -> GeckoPingProtocolHandler:
        """Build a PING response class."""
        return GeckoPingProtocolHandler(
            content=b"".join([PING_VERB, b"\x00"]), **kwargs
        )

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Check if this class can handle the packet bytes."""
        return received_bytes.startswith(PING_VERB)

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the packet bytes."""
        remainder = received_bytes[5:]
        if len(remainder) > 0:
            self._sequence = struct.unpack(">B", remainder)[0]



"""GeckoAsyncSpa class."""

import asyncio
import importlib
import logging
import socket
import time
from datetime import UTC, datetime
from functools import partial
from types import ModuleType
from typing import Any

from geckolib.driver.accessor import GeckoStructAccessor
from geckolib.driver.protocol.reminders import GeckoReminderType
from geckolib.driver.protocol.watercare import GeckoSetWatercareModeProtocolHandler

from .async_spa_descriptor import GeckoAsyncSpaDescriptor
from .async_taskman import GeckoAsyncTaskMan
from .config import (
    GeckoConfig,
    config_sleep,
    release_config_change_waiters,
    set_config_mode,
)
from .const import GeckoConstants
from .driver import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoAsyncStructure,
    GeckoAsyncUdpProtocol,
    GeckoConfigFileProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoGetWatercareModeProtocolHandler,
    GeckoPackCommandProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoRFErrProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoUnhandledProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoWatercareErrorHandler,
    Observable,
)
from .spa_events import GeckoSpaEvent

_LOGGER = logging.getLogger(__name__)


class GeckoAsyncSpa(Observable):
    """
    GeckoAsyncSpa class.

    Manage an instance of a spa, and is the main point of contact for control and
    monitoring. Uses the declarations found in pack/* to build an object that
    exposes the properties and capabilities of your spa. This class should only
    be used via a Facade which hides the implementation details.
    """

    def __init__(
        self,
        client_id: bytes,
        spa_descriptor: GeckoAsyncSpaDescriptor,
        taskman: GeckoAsyncTaskMan,
        event_handler: GeckoSpaEvent.CallBack,
    ) -> None:
        """Initialize the spa class."""
        super().__init__()

        self.client_id = client_id
        self.descriptor = spa_descriptor
        self._taskman = taskman
        self._event_handler: GeckoSpaEvent.CallBack = event_handler

        self._con_lost: asyncio.Event = asyncio.Event()
        self._transport: asyncio.BaseTransport | None = None
        self._protocol: GeckoAsyncUdpProtocol | None = None
        self._is_connected = False

        self.intouch_version_en = ""
        self.intouch_version_co = ""

        self.channel = 0
        self.signal = 0

        self.config_version = 0
        self.log_version = 0

        self.pack_type = ""

        self.pack = None
        self.version = ""
        self.config_number = 0

        self.struct: GeckoAsyncStructure = GeckoAsyncStructure(self.async_on_set_value)
        self._last_ping_at: datetime | None = None
        self._needs_reload: bool = False

    @property
    def sendparms(self) -> tuple:
        """Get send parms suitable for the send() API."""
        return (
            self.descriptor.destination[0],
            self.descriptor.destination[1],
            self.descriptor.identifier,
            self.client_id,
        )

    @property
    def is_connected(self) -> bool:
        """Is this spa connected."""
        return self._is_connected

    @property
    def revision(self) -> str:
        """Get the revision of the spa pack structure."""
        assert self.struct.pack_class is not None  # noqa: S101
        return self.struct.pack_class.revision

    def _get_version_handler_func(self) -> GeckoVersionProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        return GeckoVersionProtocolHandler.request(
            self._protocol.get_and_increment_sequence_counter(),
            parms=self.sendparms,
        )

    def _get_channel_handler_func(self) -> GeckoGetChannelProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        return GeckoGetChannelProtocolHandler.request(
            self._protocol.get_and_increment_sequence_counter(),
            parms=self.sendparms,
        )

    def _get_config_file_handler_func(self) -> GeckoConfigFileProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        return GeckoConfigFileProtocolHandler.request(
            self._protocol.get_and_increment_sequence_counter(),
            parms=self.sendparms,
        )

    async def _async_on_rferr(
        self, handler: GeckoRFErrProtocolHandler, _sender: tuple
    ) -> None:
        await self._event_handler(GeckoSpaEvent.ERROR_RF_ERROR)
        # If we've had too many RFErr messages, then we bail waiting for user action
        if handler.total_error_count > GeckoConstants.MAX_RF_ERRORS_BEFORE_HALT:
            _LOGGER.error(
                (
                    "Spa %s has too many RF errors, so disconnecting "
                    "awaiting user intervention"
                ),
                self.descriptor.name,
            )
            await self._event_handler(GeckoSpaEvent.ERROR_TOO_MANY_RF_ERRORS)
            set_config_mode(active=False)

    async def _async_import_module(
        self, loop: asyncio.AbstractEventLoop, module: str
    ) -> ModuleType:
        return await loop.run_in_executor(
            None, partial(importlib.import_module, module)
        )

    async def _connect(self) -> None:
        loop = asyncio.get_running_loop()
        self._con_lost.clear()

        self._transport, _protocol = await loop.create_datagram_endpoint(
            lambda: GeckoAsyncUdpProtocol(
                self._taskman,
                self._con_lost,
                (self.descriptor.destination[0], self.descriptor.destination[1]),
            ),
            family=socket.AF_INET,
        )
        assert isinstance(_protocol, GeckoAsyncUdpProtocol)  # noqa: S101
        self._protocol = _protocol
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - after protocol creation",
        )

        await self._connect_install_task_handlers()

        if not await self._connect_get_version():
            return

        get_channel_handler = await self._protocol.get(self._get_channel_handler_func)
        if get_channel_handler is None:
            _LOGGER.error("Cannot get channel, protocol retry time exceeded")
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED
            )
            return

        self.channel = get_channel_handler.channel
        self.signal = get_channel_handler.signal_strength
        _LOGGER.debug("Got channel %s/%s, now get config", self.channel, self.signal)
        await self._event_handler(
            GeckoSpaEvent.CONNECTION_GOT_CHANNEL,
            channel=self.channel,
            signal=self.signal,
        )
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - after channel",
        )

        config_file_handler = await self._protocol.get(
            self._get_config_file_handler_func,
            None,
            4,
        )
        if config_file_handler is None:
            _LOGGER.error("Cannot get file, protocol retry time exceeded")
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED
            )
            return

        # Stash the config and log structure declarations
        self.config_version = config_file_handler.config_version
        self.log_version = config_file_handler.log_version
        plateform_key = config_file_handler.plateform_key.lower()

        _LOGGER.debug(
            "Got %s config version %s and log version %s, check if we can support this",
            plateform_key,
            self.config_version,
            self.log_version,
        )
        await self._event_handler(
            GeckoSpaEvent.CONNECTION_GOT_CONFIG_FILES,
            plateform_key=plateform_key,
            config_version=self.config_version,
            log_version=self.log_version,
        )
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - after files",
        )

        await self._event_handler(GeckoSpaEvent.CONNECTION_INITIAL_DATA_BLOCK_REQUEST)
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - after initial block request",
        )

        if not await self.struct.get(
            self._protocol,
            lambda: GeckoStatusBlockProtocolHandler.full_request(
                self._protocol.get_and_increment_sequence_counter(),
                parms=self.sendparms,
            ),
            5,
        ):
            _LOGGER.error("Cannot get full struct, protocol retry time exceeded")
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED
            )

            return

        if not await self._connect_load_pack(plateform_key):
            return

        await self.struct.check_for_accessories()

        _LOGGER.debug("Status block was completed, so spa can be connected")

        self.struct.build_accessors()

        self.pack = self.accessors[GeckoConstants.KEY_PACK_TYPE].value
        if GeckoConstants.KEY_PACK_CORE_ID in self.accessors:
            self.version = (
                f"{self.accessors[GeckoConstants.KEY_PACK_CORE_ID].value}"
                f" v{self.accessors[GeckoConstants.KEY_PACK_CORE_REV].value}"
                f".{self.accessors[GeckoConstants.KEY_PACK_CORE_REL].value}"
            )
        else:
            self.version = (
                f"{self.accessors[GeckoConstants.KEY_PACK_CONFIG_ID].value}"
                f"v{self.accessors[GeckoConstants.KEY_PACK_CONFIG_REV].value}"
                f".{self.accessors[GeckoConstants.KEY_PACK_CONFIG_REL].value}"
            )

        if GeckoConstants.KEY_CONFIG_NUMBER in self.accessors:
            self.config_number = self.accessors[GeckoConstants.KEY_CONFIG_NUMBER].value

        self._connect_listen_too_critical_changes()

        self._is_connected = True
        self._needs_reload = False
        await self._event_handler(GeckoSpaEvent.CONNECTION_SPA_COMPLETE)
        _LOGGER.debug("Spa connected")

    async def _connect_install_task_handlers(self) -> None:
        if self._protocol is None:
            return

        self._taskman.add_task(
            GeckoUnhandledProtocolHandler().consume(self._protocol),
            "Unhandled packet",
            "SPA",
        )
        self._taskman.add_task(
            GeckoPacketProtocolHandler(async_on_handled=self._async_on_packet).consume(
                self._protocol
            ),
            "Packet handler",
            "SPA",
        )
        self._taskman.add_task(
            GeckoAsyncPartialStatusBlockProtocolHandler(
                self._protocol, async_on_handled=self._async_on_partial_status_update
            ).consume(self._protocol),
            "Partial status block handler",
            "SPA",
        )
        self._taskman.add_task(
            GeckoRFErrProtocolHandler(async_on_handled=self._async_on_rferr).consume(
                self._protocol
            ),
            "RFErr handler",
            "SPA",
        )
        self._taskman.add_task(
            GeckoWatercareErrorHandler(async_on_handled=self._async_on_wcerr).consume(
                self._protocol
            ),
            "WCErr handler",
            "SPA",
        )
        self._taskman.add_task(self._ping_loop(), "Ping loop", "SPA")
        self._taskman.add_task(self._refresh_loop(), "Refresh loop", "SPA")
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - before version ",
        )

    async def _connect_get_version(self) -> bool:
        version_handler = await self._protocol.get(self._get_version_handler_func)
        if version_handler is None:
            _LOGGER.error("Cannot get version, protocol retry time exceeded")
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED
            )
            return False

        self.intouch_version_en = (
            f"{version_handler.en_build}"
            f" v{version_handler.en_major}.{version_handler.en_minor}"
        )
        self.intouch_version_co = (
            f"{version_handler.co_build}"
            f" v{version_handler.co_major}.{version_handler.co_minor}"
        )
        _LOGGER.debug(
            "Got in.touch2 firmware version EN(Home) %s/CO(Spa) %s, now get channel",
            self.intouch_version_en,
            self.intouch_version_co,
        )
        await self._event_handler(
            GeckoSpaEvent.CONNECTION_GOT_FIRMWARE_VERSION,
            intouch_version_en=self.intouch_version_en,
            intouch_version_co=self.intouch_version_co,
        )
        await config_sleep(
            GeckoConstants.CONNECTION_STEP_PAUSE_IN_SECONDS,
            "Async spa connect - After version",
        )

        return True

    async def _connect_load_pack(self, plateform_key: str) -> bool:
        try:
            await self.struct.load_pack_class(plateform_key)
        except ModuleNotFoundError as ex:
            _LOGGER.exception("Cannot load pack %s", plateform_key)
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_CANNOT_FIND_SPA_PACK, exception=ex.args
            )
            return False

        try:
            await self.struct.load_config_module(self.config_version)
        except ModuleNotFoundError as ex:
            _LOGGER.exception("Cannot load config %d", self.config_version)
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_CANNOT_FIND_CONFIG_VERSION, exception=ex.args
            )
            return False

        try:
            await self.struct.load_log_module(self.log_version)
        except ModuleNotFoundError as ex:
            _LOGGER.exception("Cannot load log %d", self.log_version)
            await self._event_handler(
                GeckoSpaEvent.CONNECTION_CANNOT_FIND_LOG_VERSION, exception=ex.args
            )
            return False

        self.pack_type = self.struct.pack_class.plateform_type
        _LOGGER.debug(
            "Got spa configuration Type %s - CFG %s/LOG %s, now ask for initial "
            "status block",
            self.pack_type,
            self.config_version,
            self.log_version,
        )
        return True

    def _connect_listen_too_critical_changes(self) -> None:
        # Listen to critical data changes for reload
        if "PackType" in self.accessors:
            self.accessors["PackType"].watch(self._on_critical_info_changed)
        if "PackConfID" in self.accessors:
            self.accessors["PackConfID"].watch(self._on_critical_info_changed)
        if "PackConfigLib" in self.accessors:
            self.accessors["PackConfigLib"].watch(self._on_critical_info_changed)
        if "PackStatusLib" in self.accessors:
            self.accessors["PackStatusLib"].watch(self._on_critical_info_changed)
        for output in self.struct.all_outputs:
            self.accessors[output].watch(self._on_critical_info_changed)
        if "InMix-PackType" in self.accessors:
            self.accessors["InMix-PackType"].watch(self._on_critical_info_changed)
        if "InMix-NumberOfZones" in self.accessors:
            self.accessors["InMix-NumberOfZones"].watch(self._on_critical_info_changed)

    async def connect(self) -> None:
        """Wrap the connection with exception safety."""
        try:
            await self._connect()
        except Exception:
            _LOGGER.exception("Exception during spa connection")
            raise

    async def disconnect(self) -> None:
        """Disconnect the spa from the async protocol."""
        self._is_connected = False
        await self._event_handler(GeckoSpaEvent.RUNNING_SPA_DISCONNECTED)
        self.struct.reset()
        self._taskman.cancel_key_tasks("SPA")
        if self._protocol is not None:
            self._protocol.disconnect()
            self._protocol = None
        self._transport = None
        self.unwatch_all()
        self._needs_reload = False

    def _on_critical_info_changed(
        self, _sender: Any, _old_value: Any, _new_value: Any
    ) -> None:
        self._needs_reload = True
        release_config_change_waiters()

    @property
    def isopen(self) -> bool:
        """Return True if the spa has a UDP socket with the in.touch EN module."""
        if self._protocol is None:
            return False
        return self._protocol.isopen

    @property
    def is_responding_to_pings(self) -> bool:
        """Is the remote end responding to pings."""
        if self._last_ping is None:
            return False
        # Allow twice as much time as ping frequency to make sure boundary tests
        # are successful
        return (
            time.monotonic() - self._last_ping
        ) < GeckoConfig.PING_DEVICE_NOT_RESPONDING_TIMEOUT_IN_SECONDS

    async def _async_on_packet(
        self, handler: GeckoPacketProtocolHandler, _sender: tuple
    ) -> None:
        if handler.parms == self.sendparms:
            assert self._protocol is not None  # noqa: S101
            assert handler.packet_content is not None  # noqa: S101
            self._protocol.datagram_received(handler.packet_content, handler.parms)
        else:
            _LOGGER.warning(
                "Dropping packet from %s because it didn't match %s",
                handler.parms,
                self.sendparms,
            )

    @property
    def last_ping_at(self) -> datetime | None:
        """The datetime of the last successful ping response or None."""
        return self._last_ping_at

    async def _ping_loop(self) -> None:
        _LOGGER.debug("Ping loop started")

        self._last_ping = time.monotonic()

        try:
            while self.isopen:
                assert self._protocol is not None  # noqa: S101
                ping_handler = await self._protocol.get(
                    lambda: GeckoPingProtocolHandler.request(parms=self.sendparms)
                )

                if ping_handler is not None:
                    self._last_ping = time.monotonic()
                    self._last_ping_at = datetime.now(tz=UTC)
                    self._on_change()
                    await self._event_handler(
                        GeckoSpaEvent.RUNNING_PING_RECEIVED,
                        last_ping_at=self._last_ping_at,
                    )
                else:
                    await self._event_handler(
                        GeckoSpaEvent.RUNNING_PING_MISSED,
                        last_ping_at=self._last_ping_at,
                    )
                    if (
                        time.monotonic() - self._last_ping
                        > GeckoConfig.PING_DEVICE_NOT_RESPONDING_TIMEOUT_IN_SECONDS
                    ):
                        await self._event_handler(
                            GeckoSpaEvent.RUNNING_PING_NO_RESPONSE,
                            last_ping_at=self._last_ping_at,
                        )

                # Ping every couple of seconds until we have had a response
                await config_sleep(
                    2
                    if self._last_ping_at is None
                    else GeckoConfig.PING_FREQUENCY_IN_SECONDS,
                    "Async spa ping loop",
                )

        except asyncio.CancelledError:
            _LOGGER.debug("Ping loop cancelled")
            raise

        except:
            _LOGGER.exception("Exception in ping loop")
            raise

        finally:
            _LOGGER.debug("Ping loop finished")

    def _get_status_block_handler_func(self) -> GeckoStatusBlockProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        assert self.struct.log_class is not None  # noqa: S101
        return GeckoStatusBlockProtocolHandler.request(
            self._protocol.get_and_increment_sequence_counter(),
            self.struct.log_class.begin,
            self.struct.log_class.end,
            parms=self.sendparms,
        )

    async def _refresh_loop(self) -> None:
        """
        Refresh Loop.

        The refresh loop is simply to ensure that our understanding
        of the live parts of the spastruct are always up-to-date. We
        shouldn't need to do this, but this is belt and braces stuff
        """
        try:
            _LOGGER.debug("Refresh loop started")
            while self.isopen:
                time_to_doit = await config_sleep(
                    GeckoConfig.SPA_PACK_REFRESH_FREQUENCY_IN_SECONDS,
                    "Async spa refresh handler ",
                )
                if self._needs_reload:
                    await self._event_handler(GeckoSpaEvent.RUNNING_SPA_NEEDS_RELOAD)
                    continue
                if not self.is_connected:
                    continue
                if not self.is_responding_to_pings:
                    continue
                if not time_to_doit:
                    continue

                if False:
                    # Removed because I don't think this is required now that ping
                    # is back to 2 seconds. Time will tell.
                    if not await self.struct.get(
                        self._protocol,
                        self._get_status_block_handler_func,
                        5,
                    ):
                        await self._event_handler(
                            GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED
                        )

                assert self._protocol is not None  # noqa: S101
                get_channel_handler = await self._protocol.get(
                    self._get_channel_handler_func
                )
                if get_channel_handler is None:
                    _LOGGER.error("Cannot get channel, protocol retry time exceeded")
                    await self._event_handler(
                        GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED
                    )
                    continue

                self.channel = get_channel_handler.channel
                self.signal = get_channel_handler.signal_strength
                _LOGGER.debug("Refresh channel %s/%s", self.channel, self.signal)
                await self._event_handler(GeckoSpaEvent.RUNNING_SPA_PACK_REFRESHED)

        except asyncio.CancelledError:
            _LOGGER.debug("Refresh loop cancelled")
            raise

        except Exception:
            _LOGGER.exception("Refresh loop caught exception")
            raise

    async def _async_on_partial_status_update(
        self,
        handler: GeckoAsyncPartialStatusBlockProtocolHandler,
        _sender: tuple,
    ) -> None:
        for change in handler.changes:
            self.struct.replace_status_block_segment(change[0], change[1])

    async def async_on_set_value(self, pos: int, length: int, newvalue: Any) -> None:
        """Set the data block to a specific value."""
        try:
            assert self._protocol is not None  # noqa: S101
            if not self.is_connected:
                _LOGGER.warning("Cannot set value when spa not connected")
                return
            if not self.is_responding_to_pings:
                _LOGGER.debug("Cannot set value when spa not responding to pings")
                return

            pack_command_handler = await self._protocol.get(
                lambda: GeckoPackCommandProtocolHandler.set_value(
                    self._protocol.get_and_increment_sequence_counter(command=True),
                    self.pack_type,
                    self.config_version,
                    self.log_version,
                    pos,
                    length,
                    newvalue,
                    parms=self.sendparms,
                ),
                None,
                1,
            )

            if pack_command_handler is not None:
                self.struct.replace_status_block_segment(
                    pos, GeckoStructAccessor.pack_data(length, newvalue)
                )
            else:
                _LOGGER.error("Cannot set value, protocol retry time exceeded")
                await self._event_handler(
                    GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED
                )

        except asyncio.CancelledError:
            _LOGGER.debug("Async set value cancelled")
            raise
        except Exception:
            _LOGGER.exception("Async set value caught exception")
            raise

    @property
    def accessors(self) -> dict:
        """Get the accessors dictionary for this spa."""
        return self.struct.accessors

    async def async_press(self, keypad: int) -> None:
        """Simulate a button press async."""
        try:
            assert self._protocol is not None  # noqa: S101
            if not self.is_connected:
                _LOGGER.warning("Cannot press keypad when spa not connected")
                return
            if not self.is_responding_to_pings:
                _LOGGER.debug("Cannot press keypad when spa not responding to pings")
                return
            _LOGGER.info("Async press of key %d", keypad)

            pack_command_handler = await self._protocol.get(
                lambda: GeckoPackCommandProtocolHandler.keypress(
                    self._protocol.get_and_increment_sequence_counter(command=True),
                    self.pack_type,
                    keypad,
                    parms=self.sendparms,
                ),
                None,
                1,
            )

            if pack_command_handler is None:
                _LOGGER.error("Cannot press keypad, protocol retry time exceeded")
                await self._event_handler(
                    GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED
                )
        except asyncio.CancelledError:
            _LOGGER.debug("Async press cancelled")
            raise
        except Exception:
            _LOGGER.exception("Async press caught exception")
            raise

    def _get_watercare_handler_func(self) -> GeckoGetWatercareModeProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        return GeckoGetWatercareModeProtocolHandler.get(
            self._protocol.get_and_increment_sequence_counter(),
            parms=self.sendparms,
        )

    async def _async_on_wcerr(
        self, _handler: GeckoWatercareErrorHandler, _sender: tuple
    ) -> None:
        await self._event_handler(GeckoSpaEvent.RUNNING_SPA_WATER_CARE_ERROR)

    async def async_get_watercare_mode(self) -> int | None:
        """Get the watercare mode."""
        if not self.is_connected:
            _LOGGER.warning("Cannot get watercare when spa not connected")
            return 0
        if not self.is_responding_to_pings:
            _LOGGER.debug("Cannot get watercare when spa not responding to pings")
            return 0
        assert self._protocol is not None  # noqa: S101
        get_watercare_handler = await self._protocol.get(
            self._get_watercare_handler_func
        )

        if get_watercare_handler is None:
            _LOGGER.error("Cannot get watercare mode, protocol retry time exceeded")
            await self._event_handler(GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED)
            return None

        return get_watercare_handler.mode

    async def async_set_watercare_mode(self, new_mode: int) -> None:
        """Set the watercare more."""
        if not self.is_connected:
            _LOGGER.warning("Cannot set watercare when spa not connected")
            return
        if not self.is_responding_to_pings:
            _LOGGER.debug("Cannot set watercare when spa not responding to pings")
            return
        assert self._protocol is not None  # noqa: S101
        set_watercare_handler = await self._protocol.get(
            lambda: GeckoSetWatercareModeProtocolHandler.set(
                self._protocol.get_and_increment_sequence_counter(),
                new_mode,
                parms=self.sendparms,
            )
        )
        if set_watercare_handler is None:
            _LOGGER.error("Cannot set watercare mode, protocol retry time exceeded")
            await self._event_handler(GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED)
        else:
            _LOGGER.debug("Spa responded with %d", set_watercare_handler.mode)

    def _get_reminders_handler_func(self) -> GeckoRemindersProtocolHandler:
        assert self._protocol is not None  # noqa: S101
        return GeckoRemindersProtocolHandler.request(
            self._protocol.get_and_increment_sequence_counter(),
            parms=self.sendparms,
        )

    async def async_get_reminders(self) -> list[tuple]:
        """Get the reminders."""
        if not self.is_connected:
            _LOGGER.warning("Cannot get reminders when spa not connected")
            return []
        if not self.is_responding_to_pings:
            _LOGGER.debug("Cannot get reminders when spa not responding to pings")
            return []

        assert self._protocol is not None  # noqa: S101
        get_reminders_handler = await self._protocol.get(
            self._get_reminders_handler_func
        )

        if get_reminders_handler is None:
            _LOGGER.error("Cannot get reminders, protocol retry time exceeded")
            await self._event_handler(GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED)
            return []

        return get_reminders_handler.reminders

    async def async_set_reminders(
        self, reminders: list[tuple[GeckoReminderType, int]]
    ) -> None:
        """Set the reminders."""
        if not self.is_connected:
            _LOGGER.warning("Cannot set reminders when spa not connected")
            return
        if not self.is_responding_to_pings:
            _LOGGER.debug("Cannot set reminders when spa not responding to pings")
            return
        assert self._protocol is not None  # noqa: S101
        set_reminders_handler = await self._protocol.get(
            lambda: GeckoRemindersProtocolHandler.set(
                self._protocol.get_and_increment_sequence_counter(),
                reminders,
                parms=self.sendparms,
            )
        )

        if set_reminders_handler is None:
            _LOGGER.error("Cannot set reminders, protocol retry time exceeded")
            await self._event_handler(GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED)

    def get_snapshot_data(self) -> dict:
        """Get the snapshot data for this spa."""
        data = self.struct.get_snapshot_data()
        data.update(
            {
                "intouch version EN": self.intouch_version_en,
                "intouch version CO": self.intouch_version_co,
            }
        )
        return data



"""GeckoLogStruct - A class to manage the LogStruct for 'MrSteam v1'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 1

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 280

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return []

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return [
            "H2O2Err",
            "KeyStuckErr",
            "PowerFailErr",
            "Prr1Err",
            "Prr2Err",
            "SlaveH2O2Err",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "Mode": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/Mode",
                257,
                0,
                ["OFF", "ON"],
                None,
                2,
                "ALL",
            ),
            "Pause": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Controls/Pause", 257, 1, "ALL"
            ),
            "Diagnostic": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Controls/Diagnostic", 257, 2, "ALL"
            ),
            "ChromaState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/ChromaState",
                257,
                3,
                ["OFF", "ON"],
                None,
                2,
                "ALL",
            ),
            "SelectedProg": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/SelectedProg",
                261,
                None,
                ["User", "Prog1", "Prog2"],
                None,
                None,
                "ALL",
            ),
            "UserSetpoint": GeckoByteStructAccessor(
                self.struct, "LogStructure/Controls/UserSetpoint", 262, "ALL"
            ),
            "UserRuntime": GeckoWordStructAccessor(
                self.struct, "LogStructure/Controls/UserRuntime", 263, "ALL"
            ),
            "UserAroma": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Controls/UserAroma",
                265,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/Status/Hours", 256, None
            ),
            "ExternalProbe": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/ExternalProbe", 258, 0, None
            ),
            "WaterDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/WaterDetected", 2658, 1, None
            ),
            "NoKeypad": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/NoKeypad", 258, 2, None
            ),
            "NoRegulation": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/NoRegulation", 258, 3, None
            ),
            "ExpressCycle": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/ExpressCycle", 258, 5, None
            ),
            "SlaveMode": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/SlaveMode",
                259,
                1,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "SlaveHeaterState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/Status/SlaveHeaterState",
                259,
                3,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "PowerFailErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/PowerFailErr", 260, 0, None
            ),
            "Prr2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr2Err", 260, 1, None
            ),
            "Prr1Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Prr1Err", 260, 2, None
            ),
            "H2O2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/H2O2Err", 260, 3, None
            ),
            "SlaveH2O2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/SlaveH2O2Err", 260, 4, None
            ),
            "KeyStuckErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/KeyStuckErr", 260, 5, None
            ),
            "Jumper9": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper9", 266, 0, None
            ),
            "Jumper2": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper2", 266, 1, None
            ),
            "Jumper3": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper3", 266, 2, None
            ),
            "Jumper4": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper4", 266, 3, None
            ),
            "Jumper5": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper5", 266, 4, None
            ),
            "Jumper6": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper6", 266, 5, None
            ),
            "Jumper7": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper7", 266, 6, None
            ),
            "Jumper8": GeckoBoolStructAccessor(
                self.struct, "LogStructure/Status/Jumper8", 266, 7, None
            ),
            "MaxRuntime": GeckoWordStructAccessor(
                self.struct, "LogStructure/Status/MaxRuntime", 267, None
            ),
            "PackBootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootID", 269, None
            ),
            "PackBootRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRev", 271, None
            ),
            "PackBootRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRel", 272, None
            ),
            "PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackType",
                273,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                    "Color_Keypad",
                    "inYJ",
                    "MrSteam",
                ],
                None,
                None,
                None,
            ),
            "PackMemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackMemRange",
                274,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
            "PackCoreID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreID", 275, None
            ),
            "PackCoreRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRev", 277, None
            ),
            "PackCoreRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRel", 278, None
            ),
            "PackConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfigLib", 279, None
            ),
            "PackStatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackStatusLib", 280, None
            ),
        }



"""Gecko automation support for sensors."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from geckolib.driver import GeckoBoolStructAccessor
from geckolib.driver.accessor import GeckoStructAccessor

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoSensorBase(GeckoAutomationFacadeBase):
    """Base sensor allows non-accessor sensors to be implemented."""

    def __init__(
        self, facade: GeckoAsyncFacade, name: str, device_class: str | None = None
    ) -> None:
        """Initialize the sensor base."""
        super().__init__(facade, name, name.upper())
        self._device_class = device_class

    @property
    def state(self) -> Any | None:
        """The state of the sensor."""
        return None

    @property
    def unit_of_measurement(self) -> str | None:
        """The unit of measurement for the sensor, or None."""
        return None

    @property
    def device_class(self) -> str | None:
        """The device class."""
        return self._device_class

    def __repr__(self) -> str:
        """Get a string representation."""
        return f"{self.name}: {self.state}"


########################################################################################
class GeckoSensor(GeckoSensorBase):
    """
    Sensors wrapper.

    Take accessors state with extra units and device
    class properties
    """

    def __init__(
        self,
        facade: GeckoAsyncFacade,
        name: str,
        accessor: GeckoStructAccessor,
        unit_accessor: GeckoStructAccessor | str | None = None,
        device_class: str | None = None,
    ) -> None:
        """Initialize the sensor class."""
        super().__init__(facade, name, device_class)
        self._accessor = accessor
        # Bubble up change notification
        accessor.watch(self._on_change)
        self._unit_of_measurement_accessor = unit_accessor
        if isinstance(self._unit_of_measurement_accessor, GeckoStructAccessor):
            unit_accessor.watch(self._on_change)

    @property
    def state(self) -> Any:
        """The state of the sensor."""
        return self._accessor.value

    @property
    def unit_of_measurement(self) -> str | None:
        """The unit of measurement for the sensor, or None."""
        if isinstance(self._unit_of_measurement_accessor, GeckoStructAccessor):
            return self._unit_of_measurement_accessor.value
        return self._unit_of_measurement_accessor

    @property
    def accessor(self) -> GeckoStructAccessor:
        """Access the accessor member."""
        return self._accessor

    @property
    def monitor(self) -> str:
        """Get monitor string."""
        return f"{self.accessor.tag}: {self.state}"


########################################################################################
class GeckoBinarySensor(GeckoSensor):
    """Binary sensors only have two states."""

    @property
    def is_on(self) -> bool:
        """Determine if the sensor is on or not."""
        state = self.state
        if isinstance(state, bool):
            return state
        if state == "":
            return False
        return state != "OFF"

    def __repr__(self) -> str:
        """Get string representation."""
        return f"{self.name}: {self.is_on}"


########################################################################################
class GeckoErrorSensor(GeckoSensorBase):
    """Error sensor aggregates all the error keys into a comma separated text string."""

    def __init__(
        self, facade: GeckoAsyncFacade, device_class: str | None = None
    ) -> None:
        """Initialise the error sensor class."""
        super().__init__(facade, "Error Sensor", device_class)
        self._state = "No errors or warnings"

        # Listen for changes to any of the error spapack accessors
        for accessor_key in facade.spa.struct.error_keys:
            accessor = facade.spa.struct.accessors[accessor_key]
            accessor.watch(self.update_state)

        # Force initial state
        self.update_state()

    @property
    def state(self) -> str:
        """The state of the sensor."""
        return self._state

    def update_state(
        self, _sender: Any = None, _old_value: Any = None, _new_value: Any = None
    ) -> None:
        """Update the state."""
        self._state = ""

        active_errors = [
            accessor
            for accessor_key, accessor in self.facade.spa.struct.accessors.items()
            if accessor_key in self.facade.spa.struct.error_keys
            and isinstance(accessor, GeckoBoolStructAccessor)
            and accessor.value is True
        ]

        if active_errors:
            self._state = ", ".join(err.tag for err in active_errors)
            _LOGGER.debug("Error sensor state is %s", self.state)
        else:
            self._state = "None"

        self._on_change(None, None, None)



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoPackCommandProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoPackCommandHandlers(unittest.TestCase):
    """Pack Command tests."""

    def test_send_construct_key_press(self) -> None:
        handler = GeckoPackCommandProtocolHandler.keypress(1, 6, 1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SPACK\x01\x06\x02\x39\x01</DATAS></PACKT>",
        )

    def test_send_construct_set_value(self) -> None:
        handler = GeckoPackCommandProtocolHandler.set_value(
            1, 6, 9, 9, 15, 2, 702, parms=PARMS
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SPACK\x01\x06\x07\x46\x09\x09\x00\x0f\x02\xbe</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoPackCommandProtocolHandler.response(parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>PACKS</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoPackCommandProtocolHandler()
        self.assertTrue(handler.can_handle(b"SPACK", PARMS))
        self.assertTrue(handler.can_handle(b"PACKS", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_key_press(self) -> None:
        handler = GeckoPackCommandProtocolHandler()
        handler.handle(b"SPACK\x01\x06\x02\x39\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)
        self.assertTrue(handler.is_key_press)
        self.assertEqual(handler.keycode, 1)
        self.assertFalse(handler.is_set_value)

    def test_recv_handle_set_value(self) -> None:
        handler = GeckoPackCommandProtocolHandler()
        handler.handle(b"SPACK\x01\x06\x07\x46\x09\x09\x00\x0f\x02\xbe", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)
        self.assertFalse(handler.is_key_press)
        self.assertTrue(handler.is_set_value)
        self.assertEqual(handler.position, 15)
        self.assertEqual(handler.new_data, b"\x02\xbe")

    def test_recv_handle_response(self) -> None:
        handler = GeckoPackCommandProtocolHandler()
        handler.handle(b"PACKS", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertFalse(handler.is_key_press)
        self.assertFalse(handler.is_set_value)


if __name__ == "__main__":
    unittest.main()



"""Gecko <PACKT> handlers."""

import logging
import re
from typing import Any

from geckolib.driver.udp_protocol_handler import GeckoUdpProtocolHandler

PACKET_OPEN = b"<PACKT>"
PACKET_CLOSE = b"</PACKT>"
SRCCN_OPEN = b"<SRCCN>"
SRCCN_CLOSE = b"</SRCCN>"
DESCN_OPEN = b"<DESCN>"
DESCN_CLOSE = b"</DESCN>"
DATAS_OPEN = b"<DATAS>"
DATAS_CLOSE = b"</DATAS>"

_LOGGER = logging.getLogger(__name__)


class GeckoPacketProtocolHandler(GeckoUdpProtocolHandler):
    """Handle the PACKT structure."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self._parms: tuple = kwargs.get("parms")
        self._content: bytes = kwargs.get("content")
        self._on_get_parms = kwargs.get("on_get_parms")
        if self._content is not None and not isinstance(self._content, bytes):
            raise TypeError(self._content, "Content must be of type `bytes`")
        self._packet_content = None
        self._sequence = None

    @property
    def send_bytes(self) -> bytes:
        """Generate the bytes to send."""
        parms = self.parms
        return b"".join(
            [
                PACKET_OPEN,
                SRCCN_OPEN,
                parms[3],
                SRCCN_CLOSE,
                DESCN_OPEN,
                parms[2],
                DESCN_CLOSE,
                DATAS_OPEN,
                self._content,
                DATAS_CLOSE,
                PACKET_CLOSE,
            ]
        )

    @property
    def parms(self) -> tuple:
        """Delegate parms."""
        if self._on_get_parms is None:
            return self._parms
        return self._on_get_parms(self)

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we unwrap this packet."""
        # If the received bytes start with <PACKET>, we can help
        return received_bytes.startswith(PACKET_OPEN) and received_bytes.endswith(
            PACKET_CLOSE
        )

    def _extract_packet_parts(self, content: bytes) -> tuple:
        match = re.search(
            b"".join(
                [
                    SRCCN_OPEN,
                    b"(.*)",
                    SRCCN_CLOSE,
                    DESCN_OPEN,
                    b"(.*)",
                    DESCN_CLOSE,
                    DATAS_OPEN,
                    b"(.*)",
                    DATAS_CLOSE,
                ]
            ),
            content,
            re.DOTALL,
        )
        if match:
            return match.groups()
        return (None, None, None)

    def handle(self, received_bytes: bytes, sender: tuple) -> None:
        """Unwrap the packet and dispatch."""
        (
            src_identifier,
            dest_identifier,
            self._packet_content,
        ) = self._extract_packet_parts(received_bytes[7:-8])
        self._parms = (sender[0], sender[1], src_identifier, dest_identifier)

    @property
    def packet_content(self) -> bytes | None:
        """Get packet content."""
        return self._packet_content

    def __repr__(self) -> str:
        """Get string representation."""
        return (
            f"{super().__repr__()}(parms={self.parms!r},"
            f" content={self.packet_content!r})"
        )



"""Snapshot helper."""

from __future__ import annotations

import ast
import logging
import re
from datetime import UTC, datetime
from pathlib import Path
from typing import Any

from geckolib.const import GeckoConstants
from geckolib.driver import GeckoStatusBlockProtocolHandler

_LOGGER = logging.getLogger(__name__)


class GeckoSnapshot:
    """Snapshot helper class."""

    def __init__(self) -> None:
        """Initialize the snapshot."""
        self._lines = []
        self._name = None
        self._timestamp = datetime.now(tz=UTC).strftime("%Y-%m-%d %H:%M:%S")
        self._pack_type: str = ""
        self._pack_conf_id = None
        self._pack_conf_rev = None
        self._pack_conf_rel = None
        self._intouch_CO = ()
        self._intouch_EN = ()
        self._config_version = None
        self._log_version = None
        self._bytes = b""
        self._status_block_handler = GeckoStatusBlockProtocolHandler()
        self._status_block_segments = []
        self._spa_name = None

        self._funcs = [
            #
            #   Snapshot set
            #
            # Match "2020-12-08 19:53:28,310 geckolib.utils.shell INFO Snapshot (Heating)", # noqa: E501
            (r"(\d+-\d+-\d+\s+\d+:\d+:\d+).*Snapshot \((.*)\)", self._re_snapshot),
            # Match "INFO:geckolib.utils.shell:Snapshot (EconomyModeActive)",
            (r"Snapshot \((.*)\)", self._re_snapshot_alt),
            # Match "Spa pack inXM 186 v3.0"
            (r"Spa pack (.*) (\d+) v(\d+)\.(\d+)", self._re_spa_pack),
            # Match "intouch version EN 88 v15.0"
            (r"intouch version EN (\d+) v(\d+)\.(\d+)", self._re_intouch_en),
            # Match "intouch version CO 89 v11.0"
            (r"intouch version CO (\d+) v(\d+)\.(\d+)", self._re_intouch_co),
            # Match "Config version 9"
            (r"Config version (\d+)", self._re_config_version),
            # Match "Log version 9"
            (r"Log version (\d+)", self._re_log_version),
            # Match "['0x5', '0x1', ... '0x0']"
            (r"\[([0-9A-Fa-fx\\' ,]*)\]", self._re_data),
            #
            #   Connection set
            #
            # Match "PackType adjusted data = inYT"
            (r"PackType adjusted data = (\w+)", self._re_spa_pack_type),
            # Match "PackConfID @ 297, Word raw data = 163"
            (r"PackConfID @ 297, Word raw data = (\d+)", self._re_spa_pack_id),
            # Match "PackConfRev @ 299, Byte raw data = 4"
            (r"PackConfRev @ 299, Byte raw data = (\d+)", self._re_spa_pack_rev),
            # Match "PackConfRel @ 300, Byte raw data = 0"
            (r"PackConfRel @ 300, Byte raw data = (\d+)", self._re_spa_pack_rel),
            # Match "Got software version 70 v14.0/69 v11.0"
            (
                r"Got software version (\d+) v(\d+).(\d+)/(\d+) v(\d+).(\d+)",
                self._re_software_version,
            ),
            # Match "Got spa configuration Type 10 - CFG 61/LOG 61"
            (
                r"Got spa configuration Type (\d+) - CFG (\d+)/LOG (\d+)",
                self._re_config_and_log,
            ),
            # Match "STATV\x15\x16'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00</DATAS>"  # noqa: E501
            (r"(STATV.*)</DATAS>", self._re_data_segment),
            # Match "INFO SNAPSHOT ========{ ... }========"
            (r"SNAPSHOT ========({.*})========", self._re_json_liine),
        ]

    def _re_snapshot(self, groups: tuple[str | Any, ...]) -> None:
        (self._timestamp, self._name) = groups

    def _re_snapshot_alt(self, groups: tuple[str | Any, ...]) -> None:
        (self._name,) = groups

    def _re_spa_pack(self, groups: tuple[str | Any, ...]) -> None:
        (
            self._pack_type,
            self._pack_conf_id,
            self._pack_conf_rev,
            self._pack_conf_rel,
        ) = groups

    def _re_spa_pack_type(self, groups: tuple[str | Any, ...]) -> None:
        (self._pack_type,) = groups

    def _re_spa_pack_id(self, groups: tuple[str | Any, ...]) -> None:
        (hex_id,) = groups
        self._pack_conf_id = f"{int(hex_id, 16)}"

    def _re_spa_pack_rev(self, groups: tuple[str | Any, ...]) -> None:
        (self._pack_conf_rev,) = groups

    def _re_spa_pack_rel(self, groups: tuple[str | Any, ...]) -> None:
        (self._pack_conf_rel,) = groups

    def _re_intouch_en(self, groups: tuple[str | Any, ...]) -> None:
        self._intouch_EN = groups

    def _re_intouch_co(self, groups: tuple[str | Any, ...]) -> None:
        self._intouch_CO = groups

    def _re_software_version(self, groups: tuple[str | Any, ...]) -> None:
        self._intouch_EN = (groups[0], groups[1], groups[2])
        self._intouch_CO = (groups[3], groups[4], groups[5])

    def _re_config_version(self, groups: tuple[str | Any, ...]) -> None:
        (self._config_version,) = groups

    def _re_log_version(self, groups: tuple[str | Any, ...]) -> None:
        (self._log_version,) = groups

    def _re_config_and_log(self, groups: tuple[str | Any, ...]) -> None:
        (type_, self._config_version, self._log_version) = groups

    def _re_data(self, groups: tuple[str | Any, ...]) -> None:
        self._bytes = bytes(
            bytearray([int(b.strip()[1:-1], 16) for b in groups[0].split(",")])
        )

    def _re_data_segment(self, groups: tuple[str | Any, ...]) -> None:
        data = groups[0].replace("'", "\\x27")
        bytes_ = ast.literal_eval(f"b'{data}'")
        self._status_block_handler.handle(bytes_, None)
        self._status_block_segments.append(self._status_block_handler.data)
        if self._status_block_handler.next == 0:
            self._bytes = b"".join(self._status_block_segments)

    def _re_json_liine(self, groups: tuple[str | Any, ...]) -> None:
        self._parse_json(groups[0])

    def parse(self, line: str) -> None:
        """Parse a snapshot line."""
        # Always bag the lines
        self._lines.append(line)

        for fn in self._funcs:
            match = re.search(fn[0], line, re.DOTALL)
            if match:
                fn[1](match.groups())

    @property
    def name(self) -> str | None:
        """Get the snapshot name."""
        return self._name

    def set_name(self, name: str) -> None:
        """Set the snapshot name."""
        self._name = name

    @property
    def timestamp(self) -> str | None:
        """Get the snapshot timestamp."""
        return self._timestamp

    @property
    def packtype(self) -> str:
        """Get the snapshot packtype."""
        if self._pack_type != "":
            return self._pack_type
        pack_bytes = self.bytes[0x1E0:0x1E4]
        pack_type = pack_bytes.decode(GeckoConstants.MESSAGE_ENCODING)
        if pack_type in GeckoConstants.PACK_NAME_ADJUSTMENTS:
            return GeckoConstants.PACK_NAME_ADJUSTMENTS[pack_type]
        return pack_type

    @property
    def spapack(self) -> str:
        """Get the spapack."""
        return (
            f"{self.packtype} {self._pack_conf_id}"
            f" v{self._pack_conf_rev}.{self._pack_conf_rel}"
        )

    @property
    def spa_name(self) -> str | None:
        """Get the name or None."""
        return self._spa_name

    @property
    def filename(self) -> str:
        """Generate snapshot filename."""
        fixed_timestamp = f"{self.timestamp}".replace(":", "_").replace(" ", "-")
        return f"{self.packtype}-{self.name}-{fixed_timestamp}.snapshot"

    @property
    def intouch_EN(self) -> tuple[int, ...]:  # noqa: N802
        """Get the intouch_EN versions."""
        return tuple(int(i) for i in self._intouch_EN)

    @property
    def intouch_EN_str(self) -> str:  # noqa: N802
        """Get the EN version as a string."""
        tup = self._intouch_EN
        if len(tup) == 3:  # noqa: PLR2004
            return f"{tup[0]} v{tup[1]}.{tup[2]}"
        return ""

    @property
    def intouch_CO(self) -> tuple[int, ...]:  # noqa: N802
        """Get the intouch_CO versions."""
        return tuple(int(i) for i in self._intouch_CO)

    @property
    def intouch_CO_str(self) -> str:  # noqa: N802
        """Get the CO version as a string."""
        tup = self.intouch_CO
        if len(tup) == 3:  # noqa: PLR2004
            return f"{tup[0]} v{tup[1]}.{tup[2]}"
        return ""

    @property
    def config_version(self) -> int:
        """Get the config version."""
        if self._config_version is not None:
            return int(self._config_version)
        return int(self.bytes[0x1E7:0x1E8].decode(GeckoConstants.MESSAGE_ENCODING))

    @property
    def log_version(self) -> int:
        """Get the log version."""
        if self._log_version is not None:
            return int(self._log_version)
        return int(self.bytes[0x1F7:0x1F8].decode(GeckoConstants.MESSAGE_ENCODING))

    @property
    def bytes(self) -> bytes:
        """Get the structure bytes."""
        return self._bytes

    def save(self, path: str) -> None:
        """Save this snapshot into the path specified."""
        with (Path(path) / Path(self.filename)).open("w") as f:
            f.writelines(self._lines)

    def __repr__(self) -> str:
        """Get string representagtion."""
        return f"{self.name} ({self.timestamp}) for {self.packtype}"

    @staticmethod
    def parse_log_file(file: str) -> list[GeckoSnapshot]:  # noqa: PLR0912
        """Parse a log file into one or more snapshots."""
        snapshots = []

        snapshot = None
        connection = None

        if file.endswith(".json"):
            with Path(file).open() as f:
                json = "".join(f.readlines())
                snapshot = GeckoSnapshot.parse_json(json)
        else:
            with Path(file).open() as f:
                for line in f:
                    if line.startswith("{"):
                        snapshot = GeckoSnapshot.parse_json(line)
                    elif "Snapshot" in line:
                        snapshot = GeckoSnapshot()

                    if snapshot:
                        if "INFO" in line:
                            snapshot.parse(line)
                        else:
                            snapshots.append(snapshot)
                            snapshot = None

                    if "Starting spa connection handshake..." in line:
                        connection = GeckoSnapshot()
                        connection.set_name("Connection found")
                    if connection:
                        if "Spa is connected" in line:
                            connection.parse(line)
                            snapshots.append(connection)
                            connection = None
                        else:
                            connection.parse(line)

        # Handle dangling snapshot
        if snapshot is not None:
            snapshots.append(snapshot)
        if connection is not None:
            snapshots.append(connection)

        return snapshots

    def _parse_json(self, json: str) -> None:
        snap = ast.literal_eval(json)
        if "Spa pack" in snap:
            self.parse(f"Spa pack {snap['Spa pack']}")
        self.parse(f"intouch version EN {snap['intouch version EN']}")
        self.parse(f"intouch version CO {snap['intouch version CO']}")
        if "Config version" in snap:
            self._config_version = snap["Config version"]
        if "Log version" in snap:
            self._log_version = snap["Log version"]
        self._bytes = bytes(
            bytearray([int(b.strip()[2:], 16) for b in snap["Status Block"]])
        )
        self._timestamp = datetime.fromisoformat(snap["Snapshot UTC Time"])
        if "Name" in snap:
            self._spa_name = snap["Name"]

    @staticmethod
    def parse_json(json: str) -> GeckoSnapshot:
        """Parse a JSON snapshot."""
        snapshot = GeckoSnapshot()
        snapshot._parse_json(json)
        return snapshot

    @staticmethod
    def create(pack_type: str, config_version: str, log_version: str) -> GeckoSnapshot:
        """Create a blank snapshot."""
        snapshot = GeckoSnapshot()

        snapshot.parse(f"Spa pack {pack_type} 186 v3.0")
        snapshot.parse("intouch version EN 88 v15.0")
        snapshot.parse("intouch version CO 89 v11.0")

        snapshot._config_version = config_version
        snapshot._log_version = log_version
        snapshot._bytes = bytearray(1024)

        return snapshot

    def is_compatible(self, other: GeckoSnapshot) -> bool:
        """Determine if this snapshot is compatible with the other."""
        if self.spapack != other.spapack:
            return False
        if self.packtype != other.packtype:
            return False
        if self.config_version != other.config_version:
            return False
        return self.log_version == other.log_version



"""AsyncTasks is a class to farm intenal new tasks out to an awaiter."""

import asyncio
import logging
from collections.abc import Coroutine
from typing import Self

from .config import GeckoConfig, config_sleep, release_config_change_waiters

_LOGGER = logging.getLogger(__name__)


class AsyncTasks:
    """
    Manage async tasks.

    Allow clients to have tasks in batches so that they can be controlled
    individually, or cancelled in a block.

    Can be used in a 'with AsyncTasks' pattern which will run the tasks
    in the event loop.
    """

    def __init__(self) -> None:
        """Initialize async tasks class."""
        self._tasks: list[asyncio.Task] = []

    async def __aenter__(self) -> Self:
        """Async enter, used from python's with statement."""
        _LOGGER.debug("Async start, adding tidy routine")
        self.add_task(self._tidy(), "Tidy tasks", "ASYNC")
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit, when out of scope."""
        await self.gather()

    def add_task(self, coroutine: Coroutine, name_: str, key_: str) -> asyncio.Task:
        """Add tasks to the task list."""
        taskname = f"{key_}:{name_}"
        for task in self._tasks:
            if not task.done() and task.get_name() == taskname:
                msg = f"Task {taskname} already running"
                raise RuntimeError(msg)
        _LOGGER.debug("Starting task `%s` in domain `%s`", name_, key_)
        task = asyncio.create_task(coroutine, name=taskname)
        self._tasks.append(task)
        return task

    def cancel_key_tasks(self, key_: str) -> None:
        """Cancel tasks that use the specified key."""
        for task in self._tasks:
            if task.get_name().startswith(f"{key_}:"):
                _LOGGER.debug("Cancel task %s", task)
                task.cancel()
        release_config_change_waiters()

    async def gather(self) -> None:
        """Cancel all tasks."""
        for task in self._tasks:
            _LOGGER.debug("Cancel task %s", task)
            task.cancel()
        # Wait for all tasks to complete
        try:
            _results = await asyncio.gather(*self._tasks, return_exceptions=True)
            for item in zip(self._tasks, _results, strict=False):
                _LOGGER.debug("    Task %s result `%r`", item[0].get_name(), item[1])
        except Exception:
            _LOGGER.exception("AsyncTasks:gather caught exception")
            raise

    async def _tidy(self) -> None:
        try:
            while True:
                await config_sleep(
                    GeckoConfig.TASK_TIDY_FREQUENCY_IN_SECONDS, "Async task tidy"
                )
                create_new_tasks = False
                if _LOGGER.isEnabledFor(logging.DEBUG):
                    for task in self._tasks:
                        if task.done():
                            _LOGGER.debug("Tidy task %s", task)
                            create_new_tasks = True
                # Create a new task list with the currently running ones
                if create_new_tasks:
                    self._tasks = [task for task in self._tasks if not task.done()]
        except asyncio.CancelledError:
            _LOGGER.debug("Tidy loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Tidy loop caught exception")
            raise



"""Gecko AVERS/SVERS handlers."""

from __future__ import annotations

import logging
import struct
from typing import Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

AVERS_VERB = b"AVERS"
SVERS_VERB = b"SVERS"
VERSION_FORMAT = ">HBBHBB"

_LOGGER = logging.getLogger(__name__)


class GeckoVersionProtocolHandler(GeckoPacketProtocolHandler):
    """Version handler class."""

    @staticmethod
    def request(seq: int, **kwargs: Any) -> GeckoVersionProtocolHandler:
        """Generate a request."""
        return GeckoVersionProtocolHandler(
            content=b"".join([AVERS_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def response(
        intouch_EN: tuple,  # noqa: N803
        intouch_CO: tuple,  # noqa: N803
        **kwargs: Any,
    ) -> GeckoVersionProtocolHandler:
        """Generate a response."""
        return GeckoVersionProtocolHandler(
            content=b"".join(
                [
                    SVERS_VERB,
                    struct.pack(
                        VERSION_FORMAT,
                        *intouch_EN,
                        *intouch_CO,
                    ),
                ]
            ),
            **kwargs,
        )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the version handler class."""
        super().__init__(**kwargs)
        self.en_build = self.en_major = self.en_minor = None
        self.co_build = self.co_major = self.co_minor = None

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this verb."""
        return received_bytes.startswith((AVERS_VERB, SVERS_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(AVERS_VERB):
            self._sequence = struct.unpack(">B", remainder)[0]
            return
        # Otherwise must be SVERS
        (
            self.en_build,
            self.en_major,
            self.en_minor,
            self.co_build,
            self.co_major,
            self.co_minor,
        ) = struct.unpack(VERSION_FORMAT, remainder)
        self._should_remove_handler = True



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoGetChannelProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoGetChannelHandler(unittest.TestCase):
    """Channel handler tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoGetChannelProtocolHandler.request(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>CURCH\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoGetChannelProtocolHandler.response(10, 33, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>CHCUR\x0a\x21</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoGetChannelProtocolHandler()
        self.assertTrue(handler.can_handle(b"CURCH", PARMS))
        self.assertTrue(handler.can_handle(b"CHCUR", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoGetChannelProtocolHandler()
        handler.handle(b"CURCH\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoGetChannelProtocolHandler()
        handler.handle(b"CHCUR\x0a\x21", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertEqual(handler.channel, 10)
        self.assertEqual(handler.signal_strength, 33)


if __name__ == "__main__":
    unittest.main()
